# Copyright (c) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testscenarios

from tempest import config
from tempest.lib import decorators
from tempest import test

from networking_fortinet.tests.tempest_plugin.tests.scenario import base

CONF = config.CONF

load_tests = testscenarios.load_tests_apply_scenarios


class TestFortigateFWaaS(base.FWaaSScenarioTest):
    scenarios = [
        # ('without router insersion', {
        #     'router_insertion': False,
        # }),
        ('with router insersion', {
            'router_insertion': True,
        }),
    ]

    @classmethod
    def skip_checks(cls):
        super(TestFortigateFWaaS, cls).skip_checks()
        if not CONF.network.public_network_id:
            msg = ('Either project_networks_reachable must be "true", or '
                   'public_network_id must be defined.')
            raise cls.skipException(msg)

    def setUp(self):
        super(TestFortigateFWaaS, self).setUp()
        required_exts = ['fwaas', 'security-group', 'router']
        if self.router_insertion:
            required_exts.append('fwaasrouterinsertion')
        for ext in required_exts:
            if not test.is_extension_enabled(ext, 'network'):
                msg = "%s Extension not enabled." % ext
                raise self.skipException(msg)
        self._router_ids = None

    def _create_server(self, network, security_group=None):
        keys = self.create_keypair()
        kwargs = {}
        if security_group is not None:
            kwargs['security_groups'] = [{'name': security_group['name']}]
        server = self.create_server(
            key_name=keys['name'],
            networks=[{'uuid': network['id']}],
            wait_until='ACTIVE',
            **kwargs)
        return server, keys

    def _create_firewall(self, **kwargs):
        if self._router_ids is not None:
            kwargs['router_ids'] = self._router_ids
        return self.create_firewall(**kwargs)

    def _empty_policy(self, **kwargs):
        # NOTE(yamamoto): an empty policy would deny all
        fw_policy = self.create_firewall_policy(firewall_rules=[])
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
        }

    def _empty_existing_policy(self, ctx):
        self.firewall_policies_client.update_firewall_policy(
            firewall_policy_id=ctx['fw_policy']['id'], firewall_rules=[])

    def _update_policy_with_allow_rule(self, ctx):
        rules = [
            self.create_firewall_rule(
                destination_ip_address=ctx['server1_fixed_ip'],
                action="allow"),
            self.create_firewall_rule(
                source_ip_address=ctx['server1_fixed_ip'],
                action="allow"),
        ]
        rule_ids = [r['id'] for r in rules]
        self.firewall_policies_client.update_firewall_policy(
            firewall_policy_id=ctx['fw_policy']['id'],
            firewall_rules=rule_ids)
        for rule_id in rule_ids:
            self.addCleanup(
                self._remove_rule_and_wait,
                firewall_id=ctx['fw']['id'],
                firewall_policy_id=ctx['fw_policy']['id'],
                firewall_rule_id=rule_id)

    def _all_disabled_rules(self, **kwargs):
        # NOTE(yamamoto): a policy whose rules are all disabled would deny all
        fw_rule = self.create_firewall_rule(action="allow", enabled=False)
        fw_policy = self.create_firewall_policy(firewall_rules=[fw_rule['id']])
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'fw_rule': fw_rule,
        }

    def _block_ip(self, server1_fixed_ip, **kwargs):
        rules = [
            # NOTE(yamamoto): The filtering is taken place after
            # destination ip is rewritten to fixed-ip.
            self.create_firewall_rule(destination_ip_address=server1_fixed_ip,
                                      action="deny"),
            #self.create_firewall_rule(destination_ip_address=server2_fixed_ip,
            #                          action="deny"),
            self.create_firewall_rule(action="allow"),
        ]
        rule_ids = [r['id'] for r in rules]
        fw_policy = self.create_firewall_policy(firewall_rules=rule_ids)
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'server1_fixed_ip': server1_fixed_ip,
            #'server2_fixed_ip': server2_fixed_ip,
        }

    def _block_subnet(self, subnet1, **kwargs):
        rules = [
            self.create_firewall_rule(destination_ip_address=subnet1["cidr"],
                                      action="deny"),
            self.create_firewall_rule(action="allow"),
        ]
        rule_ids = [r['id'] for r in rules]
        fw_policy = self.create_firewall_policy(firewall_rules=rule_ids)
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'subnet1': subnet1["cidr"],
        }

    def _block_icmp(self, **kwargs):
        fw_rule = self.create_firewall_rule(
            protocol="icmp",
            action="deny")
        fw_rule_allow = self.create_firewall_rule(
            action="allow")
        fw_policy = self.create_firewall_policy(
            firewall_rules=[fw_rule['id'], fw_rule_allow['id']])
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'fw_rule': fw_rule,
        }

    def _block_ssh(self, **kwargs):
        fw_rule = self.create_firewall_rule(
            protocol="tcp",
            destination_port="22",
            action="deny")
        fw_rule_allow = self.create_firewall_rule(
            action="allow")
        fw_policy = self.create_firewall_policy(
            firewall_rules=[fw_rule['id'], fw_rule_allow['id']])
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'fw_rule': fw_rule,
        }

    def _block_all_with_default_allow(self, **kwargs):
        fw_rule = self.create_firewall_rule(
            action="deny")
        fw_rule_allow = self.create_firewall_rule(
            action="allow")
        fw_policy = self.create_firewall_policy(
            firewall_rules=[fw_rule['id'], fw_rule_allow['id']])
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'fw_rule': fw_rule,
        }

    def _admin_disable(self, **kwargs):
        # NOTE(yamamoto): A firewall with admin_state_up=False would block all
        fw_rule = self.create_firewall_rule(action="allow")
        fw_policy = self.create_firewall_policy(firewall_rules=[fw_rule['id']])
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'],
                                  admin_state_up=False)
        self._wait_firewall_ready(fw['id'])
        return {
            'fw': fw,
            'fw_policy': fw_policy,
            'fw_rule': fw_rule,
        }

    def _allow_ssh_and_icmp(self, ctx):
        fw_ssh_rule = self.create_firewall_rule(
            protocol="tcp",
            destination_port=22,
            action="allow")
        fw_icmp_rule = self.create_firewall_rule(
            protocol="icmp",
            action="allow")
        for rule in [fw_ssh_rule, fw_icmp_rule]:
            self.firewall_policies_client.insert_firewall_rule_in_policy(
                firewall_policy_id=ctx['fw_policy']['id'],
                firewall_rule_id=rule['id'],
                insert_before=ctx['fw_rule']['id'])
            self.addCleanup(
                self._remove_rule_and_wait,
                firewall_id=ctx['fw']['id'],
                firewall_policy_id=ctx['fw_policy']['id'],
                firewall_rule_id=rule['id'])
            self._wait_firewall_ready(ctx['fw']['id'])

    def _remove_rule_and_wait(self, firewall_id, firewall_policy_id,
                              firewall_rule_id):
        self.firewall_policies_client.remove_firewall_rule_from_policy(
            firewall_policy_id=firewall_policy_id,
            firewall_rule_id=firewall_rule_id)
        self._wait_firewall_ready(firewall_id)

    def _delete_fw(self, ctx):
        self.delete_firewall_and_wait(ctx['fw']['id'])

    def _set_admin_up(self, firewall_id, up):
        self.firewalls_client.update_firewall(firewall_id=firewall_id,
                                              admin_state_up=up)
        self._wait_firewall_ready(firewall_id=firewall_id)

    def _admin_enable(self, ctx):
        self._set_admin_up(ctx['fw']['id'], up=True)

    def _remove_rule(self, ctx):
        self._remove_rule_and_wait(
            firewall_id=ctx['fw']['id'],
            firewall_policy_id=ctx['fw_policy']['id'],
            firewall_rule_id=ctx['fw_rule']['id'])

    def _disable_rule(self, ctx):
        self.firewall_rules_client.update_firewall_rule(
            firewall_rule_id=ctx['fw_rule']['id'],
            enabled=False)
        self._wait_firewall_ready(ctx['fw']['id'])

    def _remove_router_from_fw(self, ctx):
        self.firewalls_client.update_firewall(firewall_id=ctx['fw']['id'],
                                              router_ids=[])
        self._wait_firewall_ready(ctx['fw']['id'])

    def _update_block_ssh_rule_by_port(self, ctx):
        self.firewall_rules_client.update_firewall_rule(
            firewall_rule_id=ctx['fw_rule']['id'],
            destination_port="23:25")

    def _update_block_ssh_rule_by_action(self, ctx):
        self.firewall_rules_client.update_firewall_rule(
            firewall_rule_id=ctx['fw_rule']['id'],
            action="allow")

    def _update_block_ssh_rule_by_proto(self, ctx):
        self.firewall_rules_client.update_firewall_rule(
            firewall_rule_id=ctx['fw_rule']['id'],
            protocol="udp")

    def _allow_ip(self, ctx):
        self._delete_fw(ctx)
        server1_fixed_ip = ctx['server1_fixed_ip']
        #server2_fixed_ip = ctx['server2_fixed_ip']
        rules = [
            # NOTE(yamamoto): The filtering is taken place after
            # destination ip is rewritten to fixed-ip.
            # The return traffic should be allowed regardless
            # of firewall rules.
            self.create_firewall_rule(
                destination_ip_address=server1_fixed_ip,
                action="allow"),
            #self.create_firewall_rule(
            #    destination_ip_address=server2_fixed_ip,
            #    action="allow"),
        ]
        rule_ids = [r['id'] for r in rules]
        fw_policy = self.create_firewall_policy(firewall_rules=rule_ids)
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])

    def _allow_subnet(self, ctx):
        self._delete_fw(ctx)
        subnet1 = ctx['subnet1']
        rules = [
            self.create_firewall_rule(
                destination_ip_address=subnet1,
                action="allow"),
        ]
        rule_ids = [r['id'] for r in rules]
        fw_policy = self.create_firewall_policy(firewall_rules=rule_ids)
        fw = self._create_firewall(firewall_policy_id=fw_policy['id'])
        self._wait_firewall_ready(fw['id'])

    def _confirm_allowed(self, **kwargs):
        self.check_connectivity(check_reverse_icmp_ip=self._public_gateway_ip,
                                check_reverse_curl=False,
                                **kwargs)

    def _confirm_allow_novirus(self, **kwargs):
        self.check_connectivity(check_reverse_icmp_ip=self._public_gateway_ip,
                                check_reverse_curl=True,
                                **kwargs)

    def _confirm_allowed_oneway(self, **kwargs):
        # Can ping and ssh, but can't ping back to the public gateway.
        # Same as _confirm_allowed if _public_gateway_ip is None.
        self.check_connectivity(check_reverse_icmp_ip=self._public_gateway_ip,
                                should_reverse_connect=False, **kwargs)

    def _confirm_blocked(self, **kwargs):
        self.check_connectivity(should_connect=False, **kwargs)

    def _confirm_icmp_blocked_but_tcp(self, **kwargs):
        self.check_connectivity(should_connect=False, check_ssh=False,
                                **kwargs)
        self.check_connectivity(check_icmp=False, **kwargs)

    def _confirm_ssh_blocked(self, **kwargs):
        self.check_connectivity(check_ssh=False, **kwargs)
        self.check_connectivity(should_connect=False, check_ssh=True,
                                check_icmp=False, **kwargs)

    def _create_topology(self):
        """Create a topology for testing

        +--------+             +-----------+
        |"server"|             | "subnet"  |
        |   VM   +-------------+ "network" |
        +--------+             +----+------+
                                    |
                                    | router interface port
                               +----+-----+
                               | "router" |
                               +----+-----+
                                    | router gateway port
                                    |
                                    |
                               +----+------------------+
                               | existing network      |
                               | ("public_network_id") |
                               +-----------------------+
        """
        public_network_id = CONF.network.public_network_id
        network, subnet, router = self.create_networks()
        security_group = self._create_security_group()
        server, keys = self._create_server(network,
                                           security_group=security_group)
        private_key = keys['private_key']
        server_floating_ip = self.create_floating_ip(server, public_network_id)
        fixed_ip = server['addresses'].values()[0][0]['addr']
        floating_ip = server_floating_ip['floating_ip_address']
        return fixed_ip, floating_ip, subnet, private_key, router

    def _get_public_gateway_ip(self):
        self._public_gateway_ip = None
        router = self._get_router()
        ext_gw_info = router['external_gateway_info']
        ext_fixed_ips = ext_gw_info['external_fixed_ips']
        for ip in ext_fixed_ips:
            subnet_id = ip['subnet_id']
            res = self.admin_manager.subnets_client.show_subnet(subnet_id)
            subnet = res['subnet']
            # REVISIT(yamamoto): IPv4 assumption
            if subnet['ip_version'] == 4:
                self._public_gateway_ip = subnet['gateway_ip']
                return

    def _test_firewall_basic(self, block, allow=None,
                             confirm_allowed=None, confirm_blocked=None):
        if allow is None:
            allow = self._delete_fw
        if confirm_allowed is None:
            confirm_allowed = self._confirm_allowed
        if confirm_blocked is None:
            confirm_blocked = self._confirm_blocked
        ssh_login = CONF.validation.image_ssh_user

        if self.router_insertion and CONF.network.public_router_id:
            # NOTE(yamamoto): If public_router_id is configured
            # router1 and router2 will be the same router.
            msg = "This test assumes no public_router_id configured"
            raise self.skipException(msg)

        (server1_fixed_ip, server1_floating_ip, subnet1, private_key1,
         router1) = self._create_topology()
        #server2_fixed_ip, server2_floating_ip, private_key2, router2 = \
        #    self._create_topology()
        self._get_public_gateway_ip()
        if self.router_insertion:
            # Specify the router when creating a firewall and ensures that
            # the other router (router2) is not affected by the firewall
            self._router_ids = [router1['id']]
        #    confirm_allowed2 = confirm_allowed
        #    confirm_blocked2 = confirm_blocked
        else:
            # Without router insertion, all routers should be affected
            # equally
            pass
        #    confirm_allowed2 = confirm_allowed
        #    confirm_blocked2 = confirm_blocked
        self.check_connectivity(ip_address=server1_floating_ip,
                                username=ssh_login,
                                private_key=private_key1)
        #self.check_connectivity(ip_address=server2_floating_ip,
        #                        username=ssh_login,
        #                        private_key=private_key2)
        ctx = block(server1_fixed_ip=server1_fixed_ip,
                    server1_floating_ip=server1_floating_ip,
                    subnet1=subnet1)
        #            server2_fixed_ip=server2_fixed_ip,
        #            server2_floating_ip=server2_floating_ip)
        confirm_blocked(ip_address=server1_floating_ip, username=ssh_login,
                        private_key=private_key1)
        #confirm_blocked2(ip_address=server2_floating_ip, username=ssh_login,
        #                 private_key=private_key2)
        allow(ctx)
        confirm_allowed(ip_address=server1_floating_ip, username=ssh_login,
                        private_key=private_key1)
        #confirm_allowed2(ip_address=server2_floating_ip, username=ssh_login,
        #                 private_key=private_key2)

    @test.idempotent_id('5b5f57bb-e3ca-4246-9174-bb5fe9298c5f')
    def test_firewall_block_ip(self):
        self._test_firewall_basic(block=self._block_ip, allow=self._allow_ip,
                                  confirm_allowed=self._confirm_allowed_oneway)

    @test.idempotent_id('5b5f57bb-e3ca-4246-9174-bb5fe9298c5e')
    def test_firewall_block_subnet(self):
        self._test_firewall_basic(block=self._block_subnet,
                                  allow=self._allow_subnet,
                                  confirm_allowed=self._confirm_allowed_oneway)

    @test.idempotent_id('22e23dd5-c00c-4510-87ea-4874c705a45f')
    def test_firewall_block_icmp(self):
        self._test_firewall_basic(
            block=self._block_icmp,
            confirm_blocked=self._confirm_icmp_blocked_but_tcp)

    @test.idempotent_id('5db47fa0-d3e1-49f5-b6ca-6e0fa6c397c6')
    def test_firewall_insert_rule(self):
        self._test_firewall_basic(
            block=self._block_icmp,
            allow=self._allow_ssh_and_icmp,
            confirm_blocked=self._confirm_icmp_blocked_but_tcp)

    @test.idempotent_id('20466b39-e356-4e58-bb8e-199d1172eb53')
    def test_firewall_remove_rule(self):
        self._test_firewall_basic(block=self._block_all_with_default_allow,
                                  allow=self._remove_rule,
                                  confirm_allowed=self._confirm_allow_novirus)

    @decorators.skip_because(bug="0363573")
    @test.idempotent_id('deb5874a-cc43-468e-9ac6-42b9e8a767fd')
    def test_firewall_disable_rule(self):
        self._test_firewall_basic(block=self._block_all_with_default_allow,
                                  allow=self._disable_rule)

    @test.idempotent_id('18b085f2-c63a-46b4-8764-d0e8f803ede1')
    def test_firewall_remove_router_from_fw(self):
        self._test_firewall_basic(block=self._block_all_with_default_allow,
                                  allow=self._remove_router_from_fw)

    @test.idempotent_id('18b085f2-c63a-46b4-8764-d0e8f803ede2')
    def test_firewall_update_ssh_policy_by_port(self):
        self._test_firewall_basic(block=self._block_ssh,
                                  allow=self._update_block_ssh_rule_by_port,
                                  confirm_blocked=self._confirm_ssh_blocked,
                                  confirm_allowed=self._confirm_allow_novirus)

    @test.idempotent_id('18b085f2-c63a-46b4-8764-d0e8f803ede3')
    def test_firewall_update_ssh_policy_by_action(self):
        self._test_firewall_basic(block=self._block_ssh,
                                  allow=self._update_block_ssh_rule_by_action,
                                  confirm_blocked=self._confirm_ssh_blocked,
                                  confirm_allowed=self._confirm_allow_novirus)

    @test.idempotent_id('18b085f2-c63a-46b4-8764-d0e8f803ede4')
    def test_firewall_update_ssh_policy_by_proto(self):
        self._test_firewall_basic(block=self._block_ssh,
                                  allow=self._update_block_ssh_rule_by_proto,
                                  confirm_blocked=self._confirm_ssh_blocked,
                                  confirm_allowed=self._confirm_allow_novirus)

    @test.idempotent_id('18b085f2-c63a-46b4-8764-d0e8f803ede5')
    def test_firewall_empty_policy(self):
        if not self._default_allow():
            self._test_firewall_basic(block=self._empty_policy)
        else:
            self._test_firewall_basic(block=self._block_ip,
                                      allow=self._empty_existing_policy)

    @test.idempotent_id('18b085f2-c63a-46b4-8764-d0e8f803ede6')
    def test_firewall_update_policy_with_new_rule(self):
        self._test_firewall_basic(block=self._block_ip,
                                  allow=self._update_policy_with_allow_rule,
                                  confirm_allowed=self._confirm_allow_novirus)

    @decorators.skip_because(bug="0363573")
    @test.idempotent_id('30174d2a-820b-4939-85e2-49c735f7de0c')
    def test_firewall_all_disabled_rules(self):
        self._test_firewall_basic(block=self._all_disabled_rules)

    @decorators.skip_because(bug="0363573")
    @test.idempotent_id('2233ade8-266c-4781-b052-fc0be93baa93')
    def test_firewall_admin_disable(self):
        self._test_firewall_basic(block=self._admin_disable,
                                  allow=self._admin_enable)



# Copyright 2015 Fortinet, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class TaskStatus(object):
    """Task running status.

    This is used by execution/status callback function to notify the
    task manager what's the status of current task, and also used for
    indication the final task execution result.
    NONE - Reg -> PENDING --OK--> COMPLETED
                  |              /\
                  |              |
                  |              OK
                  |              |
                  -- NOK --> ROLLBACK  -- NOK --> ERROR
    """
    NONE = 0
    PENDING = 1
    ROLLBACK = 2
    COMPLETED = 3
    ERROR = 4
    ABORT = 5



# Copyright 2015 Fortinet, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import unittest2

from networking_fortinet.api_client import client
from networking_fortinet.api_client import eventlet_request as request
from networking_fortinet.api_client import exception

E_R_CLS = request.GenericRequestEventlet.__name__


class ClientTestCase(unittest2.TestCase):
    def setUp(self):
        super(ClientTestCase, self).setUp()
        self.api = [("foobar", 443, True)]
        self.user = "admin"
        self.password = ""
        self.client = client.FortiosApiClient(
            self.api, self.user, self.password)
        self.message = {
                    "name": "ext_4093",
                    "vlanid": 4093,
                    "vdom": "root",
                    "interface": "port1",
                    "ip": "192.168.30.254 255.255.255.0"
                   }

    def tearDown(self):
        self.client = None
        super(ClientTestCase, self).tearDown()

    def test_send_request_unknown_opt(self):
        with self.assertRaises(AttributeError):
            self.client.request('UNKNOWN_OPT')

    def test_send_request_good_return(self):
        with mock.patch(__name__ + '.request.' + E_R_CLS) as MockClass:
            instance = MockClass.return_value
            resp = mock.Mock()
            resp.status = int(200)
            resp.body = '{"perfect": "body"}'
            instance.start.return_value = 'good1'
            instance.join.return_value = resp
            body = self.client.request('ADD_VLAN_INTERFACE', **self.message)
            self.assertEqual(body['perfect'], 'body')

    def test_send_request_unauthorized(self):
        with mock.patch(__name__ + '.request.' + E_R_CLS) as MockClass:
            instance = MockClass.return_value
            resp = mock.Mock()
            resp.status = int(401)
            resp.body = '{}'
            instance.start.return_value = 'good1'
            instance.join.return_value = resp
            with self.assertRaises(exception.UnAuthorizedRequest):
                self.client.request('ADD_VLAN_INTERFACE', **self.message)

    def test_send_request_error_mapping(self):
        with mock.patch(__name__ + '.request.' + E_R_CLS) as MockClass:
            instance = MockClass.return_value
            instance.start.return_value = 'good1'
            instance.join.return_value.status = int(400)
            instance.join.return_value.body = '{"Invalid Certificate": ""}'
            with self.assertRaises(exception.InvalidSecurityCertificate):
                self.client.request('ADD_VLAN_INTERFACE', **self.message)
            instance.join.return_value.body = '{"Bad Request": ""}'
            with self.assertRaises(exception.BadRequest):
                self.client.request('ADD_VLAN_INTERFACE', **self.message)



# Copyright (c) 2015 Fortinet Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests

from oslo_config import cfg

from neutron.tests import base


OK = requests.codes.ok

FGT_HOST = 'fake_host'
FGT_USR = 'admin'
FGT_PWD = ''
FGT_INT_INF = 'port2'
FGT_EXT_INF = 'port3'
FGT_NPU = False

TEST_TENANT = 'test123'
TEST_NETNAME = 'net_test123'
TEST_NETWORK = 'net_test123'
TEST_SUBNET = '4.3.2.1/24'

FGT_VLANID_FROM = 1009
FGT_VLANID_TO = 1099

TEST_ROUTER = 'router_id'


class ConfigMixin(object):

    """Mock the config for Fortinet driver and service unit tests."""

    def __init__(self):
        self.mocked_parser = None

    def set_up_mocks(self):
        # Mock the configuration file
        base.BaseTestCase.config_parse()

        cfg.CONF.set_override('service_plugins', 'router_fortinet')

        # Configure the ML2 mechanism drivers and network types
        ml2_opts = {
            'mechanism_drivers': ['fortinet', 'openvswitch'],
            'tenant_network_types': ['vlan'],
        }
        for opt, val in ml2_opts.items():
            cfg.CONF.import_opt(opt, 'neutron.plugins.ml2.config', 'ml2')
            cfg.CONF.set_override(opt, val, 'ml2')

        # Configure the ML2 type_vlan opts
        cfg.CONF.import_opt('network_vlan_ranges',
                           'neutron.plugins.ml2.drivers.type_vlan',
                           group='ml2_type_vlan')

        ml2_type_vlan_opts = {
            'vlan_ranges': ['physnet1:1000:1099'],
        }
        cfg.CONF.set_override('network_vlan_ranges',
                              ml2_type_vlan_opts['vlan_ranges'],
                              'ml2_type_vlan')
        self.vlan_ranges = ml2_type_vlan_opts['vlan_ranges']

        # Configure the Fortinet mechanism driver
        fgt_test_config = {
            'address': FGT_HOST,
            'username': FGT_USR,
            'password': FGT_PWD,
            'npu_available': FGT_NPU,
            'ext_interface': FGT_EXT_INF,
            'int_interface': FGT_INT_INF
        }
        for opt, val in fgt_test_config.items():
            cfg.CONF.set_override(opt, val, 'ml2_fortinet')
        self._fortigate = cfg.CONF.ml2_fortinet


class FakeDbContract(object):

    def __init__(self, contract_id):
        self.contract_id = contract_id



# Copyright 2015 Fortinet, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import uuid

import unittest2

from networking_fortinet.tasks import tasks


class TestTasks(unittest2.TestCase):

    def setUp(self, tasks_id=None):
        super(TestTasks, self).setUp()
        self.tasks_id = tasks_id if tasks_id else str(uuid.uuid1())
        self.tasks = tasks.Tasks(self.tasks_id)

    def tearDown(self, id=None):
        pass

    def test_register(self, **subtask):
        """
        subtask is a dictory, include two parts, func and params, it will
        be executed like subtask['func'](*subtask[params]), the following
        is a example format of subtask:
        'subtask':
                {'params': (
                    <api_client.client.FortiosApiClient object at 0x2a14a90>,
                    {'id': 2, 'vdom': 'root'}
                ),
                'func': <function wrapper at 0x2b62ed8>
                }
        """
        if not subtask:
            subtask = {
                'params': 1,
                'func': time.sleep
            }
        self.tasks.register(**subtask)
        self.assertIn(subtask, self.tasks._tasks)

    def test_register_existing_task(self, **subtask):
        """
        subtask is a dictory, include two parts, func and params, it will
        be executed like subtask['func'](*subtask[params]), the following
        is a example format of subtask:
        'subtask':
                {'params': (
                    <api_client.client.FortiosApiClient object at 0x2a14a90>,
                    {'id': 2, 'vdom': 'root'}
                ),
                'func': <function wrapper at 0x2b62ed8>
                }
        """
        if not subtask:
            subtask = {
                'params': 1,
                'func': time.sleep
            }
        self.tasks.register(**subtask)
        self.tasks.register(**subtask)
        count = self.tasks._tasks.count(subtask)
        self.assertEqual(1, count)

if __name__ == '__main__':
    unittest2.main()



# Copyright 2014 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from tempest import config
from tempest import exceptions
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions as lib_exc
from tempest import test

from networking_fortinet.tests.tempest_plugin.tests.api import base

CONF = config.CONF


class FortigateFWaaSExtensionTestJSON(base.BaseFWaaSTest):

    """
    Tests the following operations in the Neutron API using the REST client for
    Neutron:

        List firewall rules
        Create firewall rule
        Update firewall rule
        Delete firewall rule
        Show firewall rule
        List firewall policies
        Create firewall policy
        Update firewall policy
        Insert firewall rule to policy
        Remove firewall rule from policy
        Insert firewall rule after/before rule in policy
        Update firewall policy audited attribute
        Delete firewall policy
        Show firewall policy
        List firewall
        Create firewall
        Update firewall
        Delete firewall
        Show firewall
    """

    @classmethod
    def resource_setup(cls):
        super(FortigateFWaaSExtensionTestJSON, cls).resource_setup()
        if not test.is_extension_enabled('fwaas', 'network'):
            msg = "FWaaS Extension not enabled."
            raise cls.skipException(msg)

    def setUp(self):
        super(FortigateFWaaSExtensionTestJSON, self).setUp()
        self.fw_rule = self.create_firewall_rule(action="allow",
                                                 protocol="tcp")
        self.fw_policy = self.create_firewall_policy()

    def _try_delete_policy(self, policy_id):
        # delete policy, if it exists
        try:
            self.firewall_policies_client.delete_firewall_policy(policy_id)
        # if policy is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

    def _try_delete_rule(self, rule_id):
        # delete rule, if it exists
        try:
            self.firewall_rules_client.delete_firewall_rule(rule_id)
        # if rule is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

    def _try_delete_firewall(self, fw_id):
        # delete firewall, if it exists
        try:
            self.firewalls_client.delete_firewall(fw_id)
        # if firewall is not found, this means it was deleted in the test
        except lib_exc.NotFound:
            pass

        self.firewalls_client.wait_for_resource_deletion(fw_id)

    def _try_delete_router(self, rt):
        # delete router, if it exists
        try:
            self.delete_router(rt)
        except lib_exc.NotFound:
            pass

    def _wait_until_ready(self, fw_id):
        target_states = ('ACTIVE', 'CREATED')

        def _wait():
            firewall = self.firewalls_client.show_firewall(fw_id)
            firewall = firewall['firewall']
            return firewall['status'] in target_states

        if not test.call_until_true(_wait, CONF.network.build_timeout,
                                    CONF.network.build_interval):
            m = ("Timed out waiting for firewall %s to reach %s state(s)" %
                 (fw_id, target_states))
            raise exceptions.TimeoutException(m)

    @test.idempotent_id('60a20691-8496-4a2d-a54c-0d5145592624')
    def test_list_firewall_rules(self):
        # List firewall rules
        fw_rules = self.firewall_rules_client.list_firewall_rules()
        fw_rules = fw_rules['firewall_rules']
        self.assertIn((self.fw_rule['id'],
                       self.fw_rule['name'],
                       self.fw_rule['action'],
                       self.fw_rule['protocol'],
                       self.fw_rule['ip_version'],
                       self.fw_rule['enabled']),
                      [(m['id'],
                        m['name'],
                        m['action'],
                        m['protocol'],
                        m['ip_version'],
                        m['enabled']) for m in fw_rules])

    @test.idempotent_id('0e8e13fb-13f3-42fc-9e97-c016bc69645a')
    def test_create_update_delete_firewall_rule(self):
        # Create firewall rule
        body = self.firewall_rules_client.create_firewall_rule(
            name=data_utils.rand_name("fw-rule"),
            action="allow",
            protocol="tcp")
        fw_rule_id = body['firewall_rule']['id']

        # Update firewall rule
        body = self.firewall_rules_client.update_firewall_rule(fw_rule_id,
                                                               shared=True)
        self.assertTrue(body["firewall_rule"]['shared'])

        # Delete firewall rule
        self.firewall_rules_client.delete_firewall_rule(fw_rule_id)
        # Confirm deletion
        fw_rules = self.firewall_rules_client.list_firewall_rules()
        self.assertNotIn(fw_rule_id,
                         [m['id'] for m in fw_rules['firewall_rules']])

    @test.idempotent_id('fb90dfb5-bb97-4191-a33c-4de7d5553629')
    def test_show_firewall_rule(self):
        # show a created firewall rule
        fw_rule = self.firewall_rules_client.show_firewall_rule(
            self.fw_rule['id'])
        for key, value in six.iteritems(fw_rule['firewall_rule']):
            self.assertEqual(self.fw_rule[key], value)

    @test.idempotent_id('5fcf7de9-e2b2-4de8-8b8f-d6affd5adc92')
    def test_list_firewall_policies(self):
        fw_policies = self.firewall_policies_client.list_firewall_policies()
        fw_policies = fw_policies['firewall_policies']
        self.assertIn((self.fw_policy['id'],
                       self.fw_policy['name'],
                       self.fw_policy['firewall_rules']),
                      [(m['id'],
                        m['name'],
                        m['firewall_rules']) for m in fw_policies])

    @test.idempotent_id('206059c6-a574-40c2-820c-81474c599274')
    def test_create_update_delete_firewall_policy(self):
        # Create firewall policy
        body = self.firewall_policies_client.create_firewall_policy(
            name=data_utils.rand_name("fw-policy"))
        fw_policy_id = body['firewall_policy']['id']
        self.addCleanup(self._try_delete_policy, fw_policy_id)

        # Update firewall policy
        body = self.firewall_policies_client.update_firewall_policy(
            fw_policy_id,
            shared=True,
            name="updated_policy")
        updated_fw_policy = body["firewall_policy"]
        self.assertTrue(updated_fw_policy['shared'])
        self.assertEqual("updated_policy", updated_fw_policy['name'])

        # Delete firewall policy
        self.firewall_policies_client.delete_firewall_policy(fw_policy_id)
        # Confirm deletion
        fw_policies = self.firewall_policies_client.list_firewall_policies()
        fw_policies = fw_policies['firewall_policies']
        self.assertNotIn(fw_policy_id, [m['id'] for m in fw_policies])

    @test.idempotent_id('03189579-c485-412c-b033-4b12ab715b6e')
    def test_show_firewall_policy(self):
        # show a created firewall policy
        fw_policy = self.firewall_policies_client.show_firewall_policy(
            self.fw_policy['id'])
        fw_policy = fw_policy['firewall_policy']
        for key, value in six.iteritems(fw_policy):
            self.assertEqual(self.fw_policy[key], value)

    @test.idempotent_id('b496a8ff-7abc-4a3f-96d9-1acbbcde2ba9')
    def test_create_show_delete_firewall(self):
        # Create tenant network resources required for an ACTIVE firewall
        network = self.create_network()
        subnet = self.create_subnet(network)
        router = self.create_router(
            data_utils.rand_name('router-'),
            admin_state_up=True)
        self.addCleanup(self._try_delete_router, router)
        self.routers_client.add_router_interface(router['id'],
                                                 subnet_id=subnet['id'])

        # Create firewall
        body = self.firewalls_client.create_firewall(
            name=data_utils.rand_name("firewall"),
            firewall_policy_id=self.fw_policy['id'])
        created_firewall = body['firewall']
        firewall_id = created_firewall['id']
        self.addCleanup(self._try_delete_firewall, firewall_id)

        # Wait for the firewall resource to become ready
        self._wait_until_ready(firewall_id)

        # show a created firewall
        firewall = self.firewalls_client.show_firewall(firewall_id)
        firewall = firewall['firewall']

        for key, value in six.iteritems(firewall):
            if key == 'status':
                continue
            self.assertEqual(created_firewall[key], value)

        # list firewall
        firewalls = self.firewalls_client.list_firewalls()
        firewalls = firewalls['firewalls']
        self.assertIn((created_firewall['id'],
                       created_firewall['name'],
                       created_firewall['firewall_policy_id']),
                      [(m['id'],
                        m['name'],
                        m['firewall_policy_id']) for m in firewalls])

        # Delete firewall
        self.firewalls_client.delete_firewall(firewall_id)

    # fortigate plugin only allow one router per tenant
    @test.idempotent_id('4ea48c81-27c2-49fd-8c8f-bb617dc0755c')
    def test_firewall_insertion_mode_add_remove_router(self):
        # Create router
        router1 = self.create_router(
            data_utils.rand_name('router-'),
            admin_state_up=True)
        self.addCleanup(self._try_delete_router, router1)
        # Create firewall on a router1
        body = self.firewalls_client.create_firewall(
            name=data_utils.rand_name("firewall"),
            firewall_policy_id=self.fw_policy['id'],
            router_ids=[router1['id']])
        created_firewall = body['firewall']
        firewall_id = created_firewall['id']
        self.addCleanup(self._try_delete_firewall, firewall_id)

        self.assertEqual([router1['id']], created_firewall['router_ids'])

        # Wait for the firewall resource to become ready
        self._wait_until_ready(firewall_id)

        # Remove router1 from the firewall
        body = self.firewalls_client.update_firewall(
            firewall_id, router_ids=[])
        updated_firewall = body['firewall']
        self.assertNotIn(router1['id'], updated_firewall['router_ids'])
        self.assertEqual(0, len(updated_firewall['router_ids']))

    @test.idempotent_id('1b5cf8f6-cd2e-4f3f-aaa0-e12c079b6d22')
    def test_firewall_insertion_mode_one_firewall_per_router(self):
        # Create router required for an ACTIVE firewall
        router = self.create_router(
            data_utils.rand_name('router1-'),
            admin_state_up=True)
        self.addCleanup(self._try_delete_router, router)
        # Create firewall
        body = self.firewalls_client.create_firewall(
            name=data_utils.rand_name("firewall"),
            firewall_policy_id=self.fw_policy['id'],
            router_ids=[router['id']])
        created_firewall = body['firewall']
        self.addCleanup(self._try_delete_firewall, created_firewall['id'])

        # Try to create firewall with the same router
        self.assertRaisesRegexp(
            lib_exc.Conflict,
            "An object with that identifier already exists",
            self.firewalls_client.create_firewall,
            name=data_utils.rand_name("firewall"),
            firewall_policy_id=self.fw_policy['id'],
            router_ids=[router['id']])

    @test.attr(type='smoke')
    @test.idempotent_id('4ebfac1d-d521-44de-a4a0-81b79312f362')
    def test_firewall_rule_insertion_position_removal_rule_from_policy(self):
        # Create firewall rule
        body = self.firewall_rules_client.create_firewall_rule(
            name=data_utils.rand_name("fw-rule"),
            action="allow",
            protocol="tcp")
        fw_rule_id1 = body['firewall_rule']['id']
        self.addCleanup(self._try_delete_rule, fw_rule_id1)
        # Create firewall policy
        body = self.firewall_policies_client.create_firewall_policy(
            name=data_utils.rand_name("fw-policy"))
        fw_policy_id = body['firewall_policy']['id']
        self.addCleanup(self._try_delete_policy, fw_policy_id)

        # Insert rule to firewall policy
        self.firewall_policies_client.insert_firewall_rule_in_policy(
            fw_policy_id, fw_rule_id1, '', '')

        # Verify insertion of rule in policy
        self.assertIn(fw_rule_id1, self._get_list_fw_rule_ids(fw_policy_id))
        # Create another firewall rule
        body = self.firewall_rules_client.create_firewall_rule(
            name=data_utils.rand_name("fw-rule"),
            action="allow",
            protocol="icmp")
        fw_rule_id2 = body['firewall_rule']['id']
        self.addCleanup(self._try_delete_rule, fw_rule_id2)

        # Insert rule to firewall policy after the first rule
        self.firewall_policies_client.insert_firewall_rule_in_policy(
            fw_policy_id, fw_rule_id2, fw_rule_id1, '')

        # Verify the position of rule after insertion
        fw_rule = self.firewall_rules_client.show_firewall_rule(
            fw_rule_id2)

        self.assertEqual(int(fw_rule['firewall_rule']['position']), 2)
        # Remove rule from the firewall policy
        self.firewall_policies_client.remove_firewall_rule_from_policy(
            fw_policy_id, fw_rule_id2)
        # Insert rule to firewall policy before the first rule
        self.firewall_policies_client.insert_firewall_rule_in_policy(
            fw_policy_id, fw_rule_id2, '', fw_rule_id1)
        # Verify the position of rule after insertion
        fw_rule = self.firewall_rules_client.show_firewall_rule(
            fw_rule_id2)
        self.assertEqual(int(fw_rule['firewall_rule']['position']), 1)
        # Remove rule from the firewall policy
        self.firewall_policies_client.remove_firewall_rule_from_policy(
            fw_policy_id, fw_rule_id2)
        # Verify removal of rule from firewall policy
        self.assertNotIn(fw_rule_id2, self._get_list_fw_rule_ids(fw_policy_id))

        # Remove rule from the firewall policy
        self.firewall_policies_client.remove_firewall_rule_from_policy(
            fw_policy_id, fw_rule_id1)

        # Verify removal of rule from firewall policy
        self.assertNotIn(fw_rule_id1, self._get_list_fw_rule_ids(fw_policy_id))

    def _get_list_fw_rule_ids(self, fw_policy_id):
        fw_policy = self.firewall_policies_client.show_firewall_policy(
            fw_policy_id)
        return [ruleid for ruleid in fw_policy['firewall_policy']
                ['firewall_rules']]

    @test.idempotent_id('d08ce90f-b7f8-4a24-8b34-50d1b035e6c8')
    def test_update_firewall_policy_audited_attribute(self):
        # Create firewall rule
        body = self.firewall_rules_client.create_firewall_rule(
            name=data_utils.rand_name("fw-rule"),
            action="allow",
            protocol="icmp")
        fw_rule_id = body['firewall_rule']['id']
        self.addCleanup(self._try_delete_rule, fw_rule_id)
        # Create firewall policy
        body = self.firewall_policies_client.create_firewall_policy(
            name=data_utils.rand_name('fw-policy'))
        fw_policy_id = body['firewall_policy']['id']
        self.addCleanup(self._try_delete_policy, fw_policy_id)
        self.assertFalse(body['firewall_policy']['audited'])
        # Update firewall policy audited attribute to true
        self.firewall_policies_client.update_firewall_policy(fw_policy_id,
                                           audited=True)
        # Insert Firewall rule to firewall policy
        self.firewall_policies_client.insert_firewall_rule_in_policy(
            fw_policy_id, fw_rule_id, '', '')
        body = self.firewall_policies_client.show_firewall_policy(
            fw_policy_id)
        self.assertFalse(body['firewall_policy']['audited'])



# Copyright 2016 Fortinet, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math
import random
import string
import time

from networking_fortinet.api_client import exception


# generate a id with the specific prefix and the width
def uniqid(prefix='', more_entropy=False, width=11):
    width = width - len(prefix)
    if width <= 0:
        raise ValueError("The width '%(w)s' is too short to generate a id "
                         "with the prefix '%(p)s'" %
                         {'w': width, 'p': prefix})
    if not more_entropy:
        m = time.time()
        uniqid = '{num1:x}{num2:x}'.format(
            num1=int(math.floor(m)), num2=int((m - math.floor(m)) * 1000000))
    else:
        valid_chars = list(set(string.hexdigits.lower()))
        uniqid = ''
        for i in range(0, width, 1):
            uniqid += random.choice(valid_chars)
    if len(uniqid) >= width:
        uniqid = prefix + uniqid[len(uniqid) - width:]
    else:
        uniqid = prefix + uniqid.zfill(width)
    return uniqid


class Prepare_vdom(object):
    def __init__(self, apiclient, vdom=None):
        self.apiclient = apiclient
        self.vdom = vdom or uniqid(prefix='tst_', more_entropy=True)

    def __enter__(self):
        try:
            self.apiclient.request("GET_VDOM", name=self.vdom)
        except exception.ResourceNotFound:
            self.apiclient.request("ADD_VDOM", name=self.vdom)
        except Exception:
            raise EnvironmentError("Create vdom %(id)s failed" %
                                   {'id': self.vdom})
        return self

    def __exit__(self, type, value, traceback):
        try:
            if self.vdom not in ['root']:
                self.apiclient.request("DELETE_VDOM", name=self.vdom)
        except Exception:
            raise EnvironmentError("Delete vdom %(id)s failed" %
                                   {'id': self.vdom})
        return self



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""create fortigate tables

Revision ID: 38462bfd9dc
Revises: None
Create Date: 2015-12-01 14:26:02.305436

"""

# revision identifiers, used by Alembic.
revision = '38462bfd9dc'
down_revision = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('fortinet_fw_rule_associations')
    op.drop_table('fortinet_firewall_vips')
    op.drop_table('fortinet_firewall_policies')
    op.drop_table('fortinet_ml2_subnets')
    op.drop_table('fortinet_firewall_ippools')
    op.drop_table('fortinet_static_routers')
    op.drop_table('fortinet_interface_subips')
    op.drop_table('fortinet_floatingip_allocations')
    op.drop_table('fortinet_vlink_ip_allocations')
    op.drop_table('fortinet_vlink_vlan_allocations')
    op.drop_table('fortinet_vdom_vlinks')
    op.drop_table('fortinet_ml2_namespaces')
    op.drop_table('fortinet_firewall_addresses')
    op.drop_table('fortinet_ml2_reservedips')
    op.drop_table('fortinet_interfaces')
    ### end Alembic commands ###


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('fortinet_interfaces',
        sa.Column('name', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('vlanid', mysql.INTEGER(display_width=11),
                  autoincrement=False, nullable=True),
        sa.Column('interface', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('type', mysql.VARCHAR(length=32),
                  server_default=sa.text(u"'vlan'"), nullable=True),
        sa.Column('ip', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('secondary_ip', mysql.VARCHAR(length=11),
                  server_default=sa.text(u"'enable'"), nullable=True),
        sa.Column('alias', mysql.VARCHAR(length=32), nullable=True),
        sa.PrimaryKeyConstraint('name'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_ml2_reservedips',
        sa.Column('port_id', mysql.VARCHAR(length=36), nullable=False),
        sa.Column('subnet_id', mysql.VARCHAR(length=36), nullable=False),
        sa.Column('mac', mysql.VARCHAR(length=32), nullable=False),
        sa.Column('ip', mysql.VARCHAR(length=32), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('edit_id', mysql.INTEGER(display_width=11),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('port_id'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_firewall_addresses',
        sa.Column('name', mysql.VARCHAR(length=40),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('subnet', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('associated_interface', mysql.VARCHAR(length=11),
                  nullable=True),
        sa.Column('group', mysql.VARCHAR(length=32), nullable=True),
        sa.PrimaryKeyConstraint('name', 'vdom'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_ml2_namespaces',
        sa.Column('id', mysql.INTEGER(display_width=11), nullable=False),
        sa.Column('tenant_id', mysql.VARCHAR(length=36), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.PrimaryKeyConstraint('id', 'tenant_id'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_vdom_vlinks',
        sa.Column('name', mysql.VARCHAR(length=11)),
        sa.Column('vdom', mysql.VARCHAR(length=11)),
        sa.PrimaryKeyConstraint('name', 'vdom'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_vlink_vlan_allocations',
        sa.Column('id', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('inf_name_int_vdom', mysql.VARCHAR(length=11),
                  nullable=True),
        sa.Column('inf_name_ext_vdom', mysql.VARCHAR(length=11),
                  nullable=True),
        sa.Column('vlanid', mysql.INTEGER(display_width=11)),
        sa.Column('allocated', mysql.TINYINT(display_width=1),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_vlink_ip_allocations',
        sa.Column('vlink_ip_subnet', mysql.VARCHAR(length=32),
                  nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('vlink_id', mysql.VARCHAR(length=36), nullable=True),
        sa.Column('allocated', mysql.TINYINT(display_width=1),
                  autoincrement=False, nullable=True),
        sa.ForeignKeyConstraint(['vlink_id'],
                                ['fortinet_vlink_vlan_allocations.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('vlink_ip_subnet'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_floatingip_allocations',
        sa.Column('ip_subnet', mysql.VARCHAR(length=32),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('floating_ip_address', mysql.VARCHAR(length=36),
                  nullable=True),
        sa.Column('vip_name', mysql.VARCHAR(length=50), nullable=True),
        sa.Column('allocated', mysql.TINYINT(display_width=1),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('ip_subnet'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_interface_subips',
        sa.Column('ip', mysql.VARCHAR(length=32),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('name', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.PrimaryKeyConstraint('ip'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_static_routers',
        sa.Column('id', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('subnet_id', mysql.VARCHAR(length=36), nullable=True),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('dst', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('device', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('gateway', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('edit_id', mysql.INTEGER(display_width=11),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_firewall_ippools',
        sa.Column('name', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('startip', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('endip', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('type', mysql.VARCHAR(length=32),
                  server_default=sa.text(u"'one-to-one'"), nullable=True),
        sa.Column('comments', mysql.VARCHAR(length=32), nullable=True),
        sa.PrimaryKeyConstraint('name'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_ml2_subnets',
        sa.Column('subnet_id', mysql.VARCHAR(length=36), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('interface', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('gateway', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('netmask', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('start_ip', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('end_ip', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('edit_id', mysql.INTEGER(display_width=11),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('subnet_id'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_firewall_policies',
        sa.Column('id', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('srcintf', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('dstintf', mysql.VARCHAR(length=11), nullable=True),
        sa.Column('srcaddr', mysql.VARCHAR(length=40),
                  server_default=sa.text(u"'all'"), nullable=True),
        sa.Column('dstaddr', mysql.VARCHAR(length=40),
                  server_default=sa.text(u"'all'"), nullable=True),
        sa.Column('poolname', mysql.VARCHAR(length=32), nullable=True),
        sa.Column('nat', mysql.VARCHAR(length=7),
                  server_default=sa.text(u"'disable'"), nullable=True),
        sa.Column('action', mysql.VARCHAR(length=11),
                  server_default=sa.text(u"'accept'"), nullable=True),
        sa.Column('service', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"'ALL'"), nullable=True),
        sa.Column('match_vip', mysql.VARCHAR(length=7),
                  server_default=sa.text(u"'disable'"), nullable=True),
        sa.Column('status', mysql.VARCHAR(length=7),
                  server_default=sa.text(u"'enable'"), nullable=True),
        sa.Column('av_profile', mysql.VARCHAR(length=35), nullable=True),
        sa.Column('webfilter_profile', mysql.VARCHAR(length=35),
                  nullable=True),
        sa.Column('ips_sensor', mysql.VARCHAR(length=35), nullable=True),
        sa.Column('application_list', mysql.VARCHAR(length=35), nullable=True),
        sa.Column('ssl_ssh_profile', mysql.VARCHAR(length=35), nullable=True),
        sa.Column('comments', mysql.VARCHAR(length=255), nullable=True),
        sa.Column('edit_id', mysql.INTEGER(display_width=11),
                  autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_firewall_vips',
        sa.Column('name', mysql.VARCHAR(length=36), nullable=False),
        sa.Column('vdom', mysql.VARCHAR(length=32), nullable=False),
        sa.Column('extip', mysql.VARCHAR(length=32), nullable=False),
        sa.Column('extintf', mysql.VARCHAR(length=32), nullable=False),
        sa.Column('mappedip', mysql.VARCHAR(length=32), nullable=True),
        sa.PrimaryKeyConstraint('name', 'vdom'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    op.create_table('fortinet_fw_rule_associations',
        sa.Column('fwr_id', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('fortinet_pid', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('type', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=False),
        sa.Column('floatingip_id', mysql.VARCHAR(length=36),
                  server_default=sa.text(u"''"), nullable=True),
        sa.ForeignKeyConstraint(['floatingip_id'], ['floatingips.id'],
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('fwr_id', 'fortinet_pid'),
        mysql_default_charset=u'utf8',
        mysql_engine=u'InnoDB'
    )
    ### end Alembic commands ###



# Copyright 2015 Fortinet, Inc.
#
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

import abc
import copy
import eventlet
try:
    import httplib as httpclient
except ImportError:
    from http import client as httpclient
import time

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
import six
import six.moves.urllib.parse as urlparse

from networking_fortinet._i18n import _LI, _LW
from networking_fortinet import api_client
from networking_fortinet.api_client import templates

LOG = logging.getLogger(__name__)

DEFAULT_HTTP_TIMEOUT = 300
DEFAULT_RETRIES = 2
DEFAULT_REDIRECTS = 2
DEFAULT_API_REQUEST_POOL_SIZE = 1
DEFAULT_MAXIMUM_REQUEST_ID = 4294967295
DOWNLOAD_TIMEOUT = 180


@six.add_metaclass(abc.ABCMeta)
class ApiRequest(object):
    '''An abstract baseclass for all ApiRequest implementations.

    This defines the interface and property structure for both eventlet and
    gevent-based ApiRequest classes.
    '''

    # List of allowed status codes.
    ALLOWED_STATUS_CODES = [
        200,
        201,
        204,
        301,
        302,
        400,
        401,
        403,
        404,
        409,
        500,
        503
    ]

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def join(self):
        pass

    def _issue_request(self):
        '''Issue a request to a provider.'''
        conn = (self._client_conn or
                self._api_client.acquire_connection(True,
                                                    copy.copy(self._headers),
                                                    rid=self._rid()))
        if conn is None:
            error = Exception(_("No API connections available"))
            self._request_error = error
            return error

        url = self._url
        LOG.debug("[%(rid)d] Issuing - request url: %(conn)s, body: %(body)s",
                  {'rid': self._rid(), 'conn': self._request_str(conn, url),
                   'body': self._body})
        issued_time = time.time()
        is_conn_error = False
        is_conn_service_unavail = False
        response = None
        try:
            redirects = 0
            while (redirects <= self._redirects):
                # Update connection with user specified request timeout,
                # the connect timeout is usually smaller so we only set
                # the request timeout after a connection is established
                if conn.sock is None:
                    conn.connect()
                    conn.sock.settimeout(self._http_timeout)
                elif conn.sock.gettimeout() != self._http_timeout:
                    conn.sock.settimeout(self._http_timeout)
                headers = copy.copy(self._headers)
                if templates.RELOGIN in url:
                    url = jsonutils.loads(templates.LOGIN)['path']
                    conn.connect()
                    self._api_client._wait_for_login(conn, headers)
                    url = self._url

                cookie = self._api_client.auth_cookie(conn)

                if (self._url != jsonutils.loads(templates.LOGIN)['path'] and
                    cookie):
                    headers['Cookie'] = cookie['Cookie']
                    headers['X-CSRFTOKEN'] = cookie['X-CSRFTOKEN']

                try:
                    if self._body:
                        if (self._url ==
                            jsonutils.loads(templates.LOGIN)['path']):
                            body = urlparse.urlencode(self._body)
                        else:
                            body = jsonutils.dumps(self._body)
                    else:
                        body = None
                    LOG.warn(_LW("Issuing request: "
                               "self._method = [%(method)s], "
                               "url=%(url)s, body=%(body)s, "
                               "headers=%(headers)s"),
                             {'method': self._method, "url": url,
                              "body": body, "headers": headers})

                    conn.request(self._method, url, body, headers)
                except Exception as e:
                    with excutils.save_and_reraise_exception():
                        LOG.warn(_LW("[%(rid)d] Exception issuing request: "
                                     "%(e)s"),
                                 {'rid': self._rid(), 'e': e})

                response = conn.getresponse()
                response.body = response.read()
                response.headers = response.getheaders()
                elapsed_time = time.time() - issued_time
                LOG.debug("@@@@@@ [ _issue_request ] [%(rid)d] "
                          "Completed request '%(conn)s': "
                          "%(status)s (%(elapsed)s seconds), "
                          "response.headers %(response.headers)s"
                          "response.body %(response.body)s",
                          {'rid': self._rid(),
                           'conn': self._request_str(conn, url),
                           'status': response.status,
                           'elapsed': elapsed_time,
                           'response.headers': response.headers,
                           'response.body': response.body})

                if response.status in (401, 302):
                    if (cookie is None and
                       self._url != jsonutils.loads(templates.LOGIN)['path']):
                        # The connection still has no valid cookie despite
                        # attempts to authenticate and the request has failed
                        # with unauthorized status code. If this isn't a
                        # a request to authenticate, we should abort the
                        # request since there is no point in retrying.
                        self._abort = True
                    LOG.debug("self._api_client=%s" % self._api_client)
                    # If request is unauthorized, clear the session cookie
                    # for the current provider so that subsequent requests
                    # to the same provider triggers re-authentication.
                    self._api_client.set_auth_cookie(conn, None)
                elif response.status == 503:
                    is_conn_service_unavail = True

                if response.status not in [301,
                                           302,
                                           307]:
                    break
                elif redirects >= self._redirects:
                    LOG.info(_LI("[%d] Maximum redirects exceeded, aborting "
                                 "request"), self._rid())
                    break
                redirects += 1
                conn, url = self._redirect_params(conn, response.headers,
                                                  self._client_conn is None)
                if url is None:
                    response.status = 500
                    break
                LOG.info(_LI("[%(rid)d] Redirecting request to: %(conn)s"),
                         {'rid': self._rid(),
                          'conn': self._request_str(conn, url)})
                # yield here, just in case we are not out of the loop yet
                eventlet.greenthread.sleep(0)
            # If we receive any of these responses, then
            # our server did not process our request and may be in an
            # errored state. Raise an exception, which will cause the
            # the conn to be released with is_conn_error == True
            # which puts the conn on the back of the client's priority
            # queue.
            if (response.status == 500 and
                response.status > 501):
                LOG.warn(_LW("[%(rid)d] Request '%(method)s %(url)s' "
                             "received: %(status)s"),
                         {'rid': self._rid(), 'method': self._method,
                          'url': self._url, 'status': response.status})
                raise Exception(_('Server error return: %s'), response.status)
            return response
        except Exception as e:
            if isinstance(e, httpclient.BadStatusLine):
                msg = ("Invalid server response")
            else:
                msg = str(e)
            if response is None:
                elapsed_time = time.time() - issued_time
            LOG.warn(_LW("[%(rid)d] Failed request '%(conn)s': '%(msg)s' "
                         "(%(elapsed)s seconds)"),
                     {'rid': self._rid(), 'conn': self._request_str(conn, url),
                      'msg': msg, 'elapsed': elapsed_time})
            self._request_error = e
            is_conn_error = True
            return e
        finally:
            # Make sure we release the original connection provided by the
            # acquire_connection() call above.
            if self._client_conn is None:
                self._api_client.release_connection(conn, is_conn_error,
                                                    is_conn_service_unavail,
                                                    rid=self._rid())

    def _redirect_params(self, conn, headers, allow_release_conn=False):
        """Process redirect response, create new connection if necessary.

        Args:
            conn: connection that returned the redirect response
            headers: response headers of the redirect response
            allow_release_conn: if redirecting to a different server,
                release existing connection back to connection pool.

        Returns: Return tuple(conn, url) where conn is a connection object
            to the redirect target and url is the path of the API request
        """
        url = None
        for name, value in headers:
            if name.lower() == "location":
                url = value
                break
        if not url:
            LOG.warn(_LW("[%d] Received redirect status without location "
                         "header field"), self._rid())
            return (conn, None)
        # Accept location with the following format:
        # 1. /path, redirect to same node
        # 2. scheme://hostname:[port]/path where scheme is https or http
        # Reject others
        # 3. e.g. relative paths, unsupported scheme, unspecified host
        result = urlparse.urlparse(url)
        if not result.scheme and not result.hostname and result.path:
            if result.path[0] == "/":
                if result.query:
                    url = "%s?%s" % (result.path, result.query)
                else:
                    url = result.path
                return (conn, url)      # case 1
            else:
                LOG.warn(_LW("[%(rid)d] Received invalid redirect location: "
                             "'%(url)s'"), {'rid': self._rid(), 'url': url})
                return (conn, None)     # case 3
        elif result.scheme not in ["http", "https"] or not result.hostname:
            LOG.warn(_LW("[%(rid)d] Received malformed redirect "
                         "location: %(url)s"),
                     {'rid': self._rid(), 'url': url})
            return (conn, None)  # case 3
        # case 2, redirect location includes a scheme
        # so setup a new connection and authenticate
        if allow_release_conn:
            self._api_client.release_connection(conn)
        conn_params = (result.hostname, result.port, result.scheme == "https")
        conn = self._api_client.acquire_redirect_connection(conn_params, True,
                                                            self._headers)
        if result.query:
            url = "%s?%s" % (result.path, result.query)
        else:
            url = result.path
        return (conn, url)

    def _rid(self):
        '''Return current request id.'''
        return self._request_id

    @property
    def request_error(self):
        '''Return any errors associated with this instance.'''
        return self._request_error

    def _request_str(self, conn, url):
        '''Return string representation of connection.'''
        return "%s %s%s" % (self._method, api_client.ctrl_conn_to_str(conn),
                            url)



# Copyright 2015 Fortinet Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import sys

from oslo_config import cfg
from oslo_log import log as logging
from neutron import version

from neutron.common import constants as l3_constants

from neutron.db import models_v2, l3_db

from neutron.db.external_net_db import ExternalNetwork

from oslo_db.sqlalchemy import session
import neutron.plugins.ml2.models as ml2_db

from networking_fortinet.common import resources
from networking_fortinet.common import utils
from networking_fortinet.ml2 import mech_fortinet
from networking_fortinet.services.l3_router import l3_fortinet
from networking_fortinet.tasks import tasks
from networking_fortinet.tasks import constants as t_consts
from networking_fortinet.db import models as fortinet_db

ROUTER_INTF = l3_constants.DEVICE_OWNER_ROUTER_INTF
ROUTER_GW = l3_constants.DEVICE_OWNER_ROUTER_GW

#streamlog = handlers.ColorHandler()
LOG = logging.getLogger(None).logger
#LOG.addHandler(streamlog)

CFG_ARGS = [
             '--config-file',
             '/etc/neutron/neutron.conf',
             '--config-file',
             '/etc/neutron/plugin.ini'
           ]

CFG_KWARGS = {}
SUPPORTED_DR = ['vlan']

cfg.CONF(args=CFG_ARGS, project='neutron',
         version='%%prog %s' % version.version_info.release_string(),
         **CFG_KWARGS)
cfg.CONF.import_group('ml2_fortinet',
                'networking_fortinet.common.config')

class Progress(object):
    def __init__(self, total, name=''):
       self.i = 0
       self.total = total if total else 1
       print "Starting %s:" % name

    def __enter__(self):
        return self

    def update(self):
        self.i += 1
        self.percent = float(self.i)/self.total
        sys.stdout.write('\r[{0:<30}] {1:.0%}'.format(
                '=' * int(round(self.percent * 29)) + '>', self.percent
            ))
        sys.stdout.flush()
        time.sleep(0.2)

    def __exit__(self, type, value, traceback):
        print ""


class Fake_context(object):
    def __init__(self, args=CFG_ARGS, kwargs=CFG_KWARGS):
        engine = session.EngineFacade.from_config(cfg.CONF)
        if not [driver for driver in cfg.CONF.ml2.type_drivers
                       if driver in SUPPORTED_DR]:
            LOG.error(_("The supported type driver %(sdr)s are not in the "
                          "ml2 type drivers %(td)s in the plugin config file.")
                        % {'sdr': SUPPORTED_DR,
                           'td': cfg.CONF.ml2.type_drivers})
            exit()
        self.session = engine.get_session(autocommit=True,
                                          expire_on_commit=False)
        self.request_id = 'migration_context'

class Fake_mech_context(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

class Fake_FortinetL3ServicePlugin(l3_fortinet.FortinetL3ServicePlugin):
    def __init__(self):
        self._fortigate = None
        self._driver = None
        self.task_manager = tasks.TaskManager()
        self.task_manager.start()
        self.Fortinet_init()

    def create_router(self, context, router):
        LOG.debug("create_router: router=%s" % (router))
        # Limit one router per tenant
        if not router.get('router', None):
            return
        tenant_id = router['router']['tenant_id']
        with context.session.begin(subtransactions=True):
            try:
                namespace = utils.add_vdom(self, context, tenant_id=tenant_id)
                utils.add_vlink(self, context, namespace.vdom)
            except Exception as e:
                LOG.error("Failed to create_router router=%(router)s",
                          {"router": router})
                resources.Exinfo(e)
                utils._rollback_on_err(self, context, e)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)


    def add_router_interface(self, context, port):
        """creates vlnk on the fortinet device."""
        db_namespace = fortinet_db.query_record(context,
                                fortinet_db.Fortinet_ML2_Namespace,
                                tenant_id=port['tenant_id'])
        vlan_inf = utils.get_intf(context, port['network_id'])
        int_intf, ext_intf = utils.get_vlink_intf(self, context,
                                       vdom=db_namespace.vdom)
        utils.add_fwpolicy(self, context,
                           vdom=db_namespace.vdom,
                           srcintf=vlan_inf,
                           dstintf=int_intf,
                           nat='enable')

    def _get_floatingip(self, context, id):
        return fortinet_db.query_record(context, l3_db.FloatingIP, id=id)


    def create_floatingip(self, context, floatingip, returned_obj):
        """Create floating IP.
        """
        LOG.debug(_("create_floatingip: floatingip=%s" % floatingip))
        self._allocate_floatingip(context, returned_obj)
        if returned_obj.get('port_id', None):
            if not floatingip['floatingip'].get('fixed_ip_address', None):
                floatingip['floatingip']['fixed_ip_address'] = \
                    returned_obj.get('fixed_ip_address')
            self._associate_floatingip(context, returned_obj['id'],
                                       floatingip)
            self.update_floatingip_status(context, returned_obj,
                                l3_constants.FLOATINGIP_STATUS_ACTIVE,
                                id=returned_obj['id'])
        return returned_obj

def init_mech_driver():
    mech_driver = mech_fortinet.FortinetMechanismDriver()
    mech_driver.initialize()
    return mech_driver

def reset(dictionary):
    for key, value in dictionary.iteritems():
        if isinstance(value, list):
            dictionary[key] = []
        elif isinstance(value, dict):
            dictionary[key] = {}
        elif isinstance(value, (int, long)):
            dictionary[key] = 0
        elif isinstance(value, (str, unicode, type(None))):
            dictionary[key] = None
        elif isinstance(value, bool):
            dictionary[key] = False
        else:
            raise TypeError
    return dictionary

def cls2dict(record, dictionary, **kwargs):
    for key, value in record.__dict__.iteritems():
        if key in dictionary:
            dictionary[key] = value
        elif key in kwargs:
            dictionary[kwargs[key]] = value
    return dictionary

def network_migration(context, mech_driver):
    """
    # networks, ml2_network_segments
    network =  {
        #'status': 'ACTIVE',
        #'subnets': [],
        'name': u'test-net',
        #'provider: physical_network': u'physnet1',
        #'admin_state_up': True,
        'tenant_id': u'11513667f4ee4a14acb0985659459988',
        'provider: network_type': u'vlan',
        'router:external': False,
        #'shared': False,
        'id': 'ff0a1d64-ce30-4ed0-ba37-597eaf8976f0',
        #'provider: segmentation_id': 1200L
    }
    # ml2_network_segments
    segments = [{
        'segmentation_id': 1200L,
        'physical_network': u'physnet1',
        'id': u'e7dfa4fb-038a-4aad-b6fa-73afba788888',
        'network_type': u'vlan'
    }]
    """
    net =  {
        'name': '',
        'tenant_id': '',
        'provider: network_type': '',
        'router:external': False,
        'id': '',
    }

    segment = {
        'segmentation_id': 0,
        'physical_network': '',
        'id': '',
        'network_type': ''
    }
    records = fortinet_db.query_records(context, models_v2.Network)
    with Progress(len(records), 'network_migration') as p:
        for record in records:
            reset(net)
            reset(segment)
            db_seg = fortinet_db.query_record(context, ml2_db.NetworkSegment,
                                              network_id=record.id)
            cls2dict(record, net)
            db_extnet = fortinet_db.query_record(context, ExternalNetwork,
                                              network_id=record.id)
            if db_extnet:
                net['router:external'] = True

            cls2dict(db_seg, segment)
            net['provider: network_type'] = db_seg.network_type
            mech_context = Fake_mech_context(_plugin_context=context,
                                             current=net,
                                             network_segments=[segment])
            mech_driver.create_network_postcommit(mech_context)
            p.update()


def subnet_migration(context, mech_driver):
    # table subnets
    subnet =  {
        'allocation_pools': [{
            'start': '172.20.21.2',
            'end': '172.20.21.254'
        }],
        'cidr': '172.20.21.0/24',
        'id': 'ee1506dc-d1a9-45b3-840e-137bdaebce52',
        'enable_dhcp': True,
        'network_id': u'ad47f7b8-4bb7-4591-b8ed-f720237dd24f',
        'tenant_id': u'11513667f4ee4a14acb0985659456f24',
        'dns_nameservers': [],
        'gateway_ip': u'172.20.21.1',
        'shared': False
    }
    ipallocation_pool = {
        'start': '172.20.21.2',
        'end': '172.20.21.254'
    }
    records = fortinet_db.query_records(context, models_v2.Subnet)
    with Progress(len(records), 'subnet_migration') as p:
        for record in records:
            dns_nameservers = []
            reset(subnet)
            reset(ipallocation_pool)
            db_ipallocation = fortinet_db.query_record(context,
                                                models_v2.IPAllocationPool,
                                                subnet_id=record.id)
            cls2dict(db_ipallocation, ipallocation_pool,
                     first_ip='start', last_ip='end')
            db_dnssrvs = fortinet_db.query_records(context,
                                                   models_v2.DNSNameServer,
                                                   subnet_id=record.id)

            for dns in db_dnssrvs:
                dns_nameservers.append(dns.address)
            cls2dict(record, subnet)
            subnet['dns_nameservers'] = dns_nameservers
            subnet['allocation_pools'] = [ipallocation_pool]
            mech_context = Fake_mech_context(_plugin_context=context,
                                             current=subnet)
            mech_driver.create_subnet_postcommit(mech_context)
            p.update()

def port_migration(context, mech_driver, l3_driver):
    """
    :param mech_driver:
    :param context:
    :return:
    # table ports
    port
    {
        'status': 'DOWN',
        'binding: host_id': '',
        'allowed_address_pairs': [],
        'device_owner': 'network: router_interface',
        'binding: profile': {

        },
        # table ipallocations
        'fixed_ips': [{
            'subnet_id': u'f645b09c-a34a-42fb-9c14-b999e43a54c7',
            'ip_address': u'172.20.21.1'
        }],
        'id': 'fb66def6-bd5e-44a0-a3f7-7c0e8e08d9ff',
        'security_groups': [],
        'device_id': u'e4020c65-7003-468b-a34d-31af297397a0',
        'name': '',
        'admin_state_up': True,
        'network_id': u'f8e34426-ccf7-429c-b726-3809d54cabdc',
        'tenant_id': u'11513667f4ee4a14acb0985659456f24',
        'binding: vif_details': {
        },
        'binding: vnic_type': 'normal',
        'binding: vif_type': 'unbound',
        'mac_address': u'00: 0c: 29: d9: 18: 3f'
    }
    """
    port = {
        'device_owner': 'network: router_interface',
        'fixed_ips': [{
            'subnet_id': u'f645b09c-a34a-42fb-9c14-b999e43a54c7',
            'ip_address': u'172.20.21.1'
        }],
        'id': 'fb66def6-bd5e-44a0-a3f7-7c0e8e08d9ff',
        'device_id': u'e4020c65-7003-468b-a34d-31af297397a0',
        'admin_state_up': True,
        'network_id': u'f8e34426-ccf7-429c-b726-3809d54cabdc',
        'tenant_id': u'11513667f4ee4a14acb0985659456f24',
        'mac_address': u'00: 0c: 29: d9: 18: 3f'
    }
    ipallocation = {
        'subnet_id': u'f645b09c-a34a-42fb-9c14-b999e43a54c7',
        'ip_address': u'172.20.21.1'
    }
    MAC = utils.get_mac(mech_driver, context)
    records = fortinet_db.query_records(context, models_v2.Port)
    with Progress(len(records), 'port_migration') as p:
         for record in records:
            reset(port)
            cls2dict(record, port)
            if port['fixed_ips']:
                fixed_ips = []
                for fixed_ip in port['fixed_ips']:
                    cls2dict(fixed_ip, ipallocation)
                    fixed_ips.append(ipallocation)
                port['fixed_ips'] = fixed_ips
            if port['device_owner'] in [ROUTER_INTF, ROUTER_GW] and \
               MAC not in port['mac_address']:
                port['mac_address'] = MAC
                if not fortinet_db.query_count(context, models_v2.Port,
                    mac_address=MAC, network_id=record.network_id):
                    fortinet_db.update_record(context, record,
                                              mac_address=MAC)
            mech_context = Fake_mech_context(_plugin_context=context,
                                             current=port)
            mech_driver.create_port_precommit(mech_context)
            mech_driver.create_port_postcommit(mech_context)
            db_routerport = fortinet_db.query_record(context,
                                                     l3_db.RouterPort,
                                                     port_id=record.id)
            if getattr(db_routerport, 'port_type', None) in [ROUTER_INTF]:
                l3_driver.add_router_interface(context, port)
            p.update()


def router_migration(context, l3_driver):
    """
    # table routers, router_extra_attributes
    router={
        u'router': {
            'external_gateway_info': None,
            u'name': u'adm_router',
            u'admin_state_up': True,
            u'tenant_id': u'01c2468ab38b4d4490a39765bb87cb00',
            'distributed': 'fakedistributed',
            'ha': 'fakeha'
        }
    }
    """
    router_obj = {
        'name': 'adm_router',
        'admin_state_up': True,
        'tenant_id': u'01c2468ab38b4d4490a39765bb87cb00'
    }
    router = {'router': router_obj}

    records = fortinet_db.query_records(context, l3_db.Router)
    with Progress(len(records), 'router_migration') as p:
        for record in records:
            reset(router_obj)
            cls2dict(record, router_obj)
            l3_driver.create_router(context, router)
            p.update()


def floatingip_migration(context, l3_driver):
    """
    # table floatingips, ipallocations
    floatingip = {
        u'floatingip': {
            u'floating_network_id': u'2bdcaa63-22c5-4e58-8e2e-8f35bef7f513',
            'tenant_id': u'11513667f4ee4a14acb0985659456f24',
            'fixed_ip_address': None,
            'port_id': None
        }
    }
    returned_obj
    {
        'floating_network_id': u'2bdcaa63-22c5-4e58-8e2e-8f35bef7f513',
        'router_id': None,
        'fixed_ip_address': None,
        'floating_ip_address': u'10.160.37.139',
        'tenant_id': u'11513667f4ee4a14acb0985659456f24',
        'status': 'DOWN',
        'port_id': None,
        'id': '78764016-da62-42fd-96a4-f2bd0510b5bc'
    }
    """

    returned_obj = {
        'fixed_ip_address': None,
        'floating_ip_address': u'10.160.37.139',
        'tenant_id': u'11513667f4ee4a14acb0985659456f24',
        'status': 'DOWN',
        'port_id': None,
        'id': '78764016-da62-42fd-96a4-f2bd0510b5bc'
        }
    floatingip = {'floatingip': returned_obj}
    records = fortinet_db.query_records(context, l3_db.FloatingIP)
    with Progress(len(records), 'floatingip_migration') as p:
        for record in records:
            reset(returned_obj)
            cls2dict(record, returned_obj, fixed_port_id='port_id')
            l3_driver.create_floatingip(context, floatingip, returned_obj)
            p.update()


def main():
    try:
        context = Fake_context()
        mech_driver = init_mech_driver()
        l3_driver = Fake_FortinetL3ServicePlugin()
        router_migration(context, l3_driver)
        network_migration(context, mech_driver)
        subnet_migration(context, mech_driver)
        port_migration(context, mech_driver, l3_driver)
        floatingip_migration(context, l3_driver)
    except Exception as e:
        raise(e)
    print "\nmigration completed.\n"


if __name__ == "__main__":
    main()




# Copyright 2015 Fortinet Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from networking_fortinet.api_client import client

ML2_FORTINET = [
    cfg.StrOpt('address', default='',
               help=_('The address of fortigates to connect to')),
    cfg.StrOpt('port', default='443',
               help=_('The FGT port to serve API requests')),
    cfg.StrOpt('protocol', default='https',
               help=_('The FGT uses which protocol: http or https')),
    cfg.StrOpt('username', default='admin',
               help=_('The username used to login')),
    cfg.StrOpt('password', default='password', secret=True,
               help=_('The password used to login')),
    cfg.StrOpt('int_interface', default='internal',
               help=_('The interface to serve tenant network')),
    cfg.StrOpt('ext_interface', default='',
               help=_('The interface to the external network')),
    cfg.StrOpt('tenant_network_type', default='vlan',
               help=_('tenant network type, default is vlan')),
    cfg.StrOpt('vlink_vlan_id_range', default='3500:4000',
               help=_('vdom link vlan interface, default is 3500:4000')),
    cfg.StrOpt('vlink_ip_range', default='169.254.0.0/20',
               help=_('vdom link interface IP range, '
                     'default is 169.254.0.0/20')),
    cfg.StrOpt('vip_mappedip_range', default='169.254.128.0/23',
               help=_('The intermediate IP range in floating IP process, '
                     'default is 169.254.128.0/23')),
    cfg.BoolOpt('npu_available', default=True,
                help=_('If npu_available is True, it requires hardware FGT'
                      'with NPU, default is True')),
    cfg.BoolOpt('enable_default_fwrule', default=False,
                help=_('If True, fwaas will add a deny all rule automatically,'
                       ' otherwise users need to add it manaully.')),
    cfg.StrOpt('av_profile', default=None,
               help=_('Assign a default antivirus profile in FWaaS, '
                     'the profile must exist in FGT, default is ""')),
    cfg.StrOpt('webfilter_profile', default=None,
               help=_('Assign a default web filter profile in FWaaS, '
                     'the profile must exist in FGT, default is ""')),
    cfg.StrOpt('ips_sensor', default=None,
               help=_('Assign a default IPS profile in FWaaS, '
                     'the profile must exist in FGT, default is ""')),
    cfg.StrOpt('application_list', default=None,
               help=_('Assign a default application control profile in FWaaS,'
                     ' the profile must exist in FGT, default is ""')),
    cfg.StrOpt('ssl_ssh_profile', default=None,
               help=_('Assign a default SSL/SSH inspection profile in FWaaS, '
                     'the profile must exist in FGT, default is ""'))
]

cfg.CONF.register_opts(ML2_FORTINET, "ml2_fortinet")

fgt_info = {
    'address': cfg.CONF.ml2_fortinet.address,
    'port': cfg.CONF.ml2_fortinet.port,
    'protocol': cfg.CONF.ml2_fortinet.protocol,
    'username': cfg.CONF.ml2_fortinet.username,
    'password': cfg.CONF.ml2_fortinet.password,
    'int_interface': cfg.CONF.ml2_fortinet.int_interface,
    'ext_interface': cfg.CONF.ml2_fortinet.ext_interface,
    'tenant_network_type': cfg.CONF.ml2_fortinet.tenant_network_type,
    'vlink_vlan_id_range': cfg.CONF.ml2_fortinet.vlink_vlan_id_range,
    'vlink_ip_range': cfg.CONF.ml2_fortinet.vlink_ip_range,
    'vip_mappedip_range': cfg.CONF.ml2_fortinet.vip_mappedip_range,
    'npu_available': cfg.CONF.ml2_fortinet.npu_available,
    'enable_default_fwrule': cfg.CONF.ml2_fortinet.enable_default_fwrule,
    'av_profile': cfg.CONF.ml2_fortinet.av_profile,
    'webfilter_profile': cfg.CONF.ml2_fortinet.webfilter_profile,
    'ips_sensor': cfg.CONF.ml2_fortinet.ips_sensor,
    'application_list': cfg.CONF.ml2_fortinet.application_list,
    'ssl_ssh_profile': cfg.CONF.ml2_fortinet.ssl_ssh_profile
}


def get_apiclient():
    """Fortinet api client initialization."""
    api_server = [(fgt_info['address'], fgt_info['port'],
                  'https' == fgt_info['protocol'])]
    return client.FortiosApiClient(
        api_server, fgt_info['username'], fgt_info['password'])



# Copyright (C) 2009-2012 VMware, Inc. All Rights Reserved.
# Copyright 2015 Fortinet, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    import httplib
except ImportError:
    import http.client as httplib
import random
import types

import eventlet
import mock
from oslo_log import log as logging
import six.moves.urllib as urllib

from neutron.tests import base

from networking_fortinet.api_client import client as fortiosclient
from networking_fortinet.api_client import eventlet_client as client
from networking_fortinet.api_client import eventlet_request as request


LOG = logging.getLogger(__name__)


REQUEST_TIMEOUT = 1

EVT_CLIENT_PATH = client.EventletApiClient.__module__ + '.EventletApiClient'


def fetch(url):
    return urllib.urlopen(url).read()


class ApiRequestEventletTestCase(base.BaseTestCase):

    def setUp(self):

        super(ApiRequestEventletTestCase, self).setUp()
        self.client = client.EventletApiClient(
            [("127.0.0.1", 80, False)], "admin", "admin")
        self.url = "/abc"
        self.req = request.EventletApiRequest(self.client, self.url)
        self.fortiosclient = fortiosclient.FortiosApiClient(
            [("127.0.0.1", 80, False)], "admin", "admin")

    def tearDown(self):
        self.client = None
        self.req = None
        self.fortiosclient = None
        super(ApiRequestEventletTestCase, self).tearDown()

    def test_construct_eventlet_api_request(self):
        e = request.EventletApiRequest(self.client, self.url)
        self.assertIsNotNone(e)

    def test_apirequest_spawn(self):
        def x(id):
            eventlet.greenthread.sleep(random.random())

        for i in range(10):
            request.EventletApiRequest._spawn(x, i)

    def test_join_with_handle_request(self):
        self.req._handle_request = mock.Mock()
        self.req.start()
        self.req.join()
        self.assertTrue(self.req._handle_request.called)

    def test_join_without_handle_request(self):
        self.req._handle_request = mock.Mock()
        self.req.join()
        self.assertFalse(self.req._handle_request.called)

    def test_request_error(self):
        self.assertIsNone(self.req.request_error)

    def test_run_and_handle_request(self):
        self.req._request_timeout = None
        self.req._handle_request = mock.Mock()
        self.req.start()
        self.req.join()
        self.assertTrue(self.req._handle_request.called)

    def test_run_and_timeout(self):
        def my_handle_request(self):
            LOG.info(_('my_handle_request() self: %s'), self)
            LOG.info(_('my_handle_request() dir(self): %s'), dir(self))
            eventlet.greenthread.sleep(REQUEST_TIMEOUT * 2)

        self.req._request_timeout = REQUEST_TIMEOUT
        self.req._handle_request = types.MethodType(
            my_handle_request, self.req)
        self.req.start()
        self.assertIsNone(self.req.join())

    def prep_issue_request(
        self, url='coolurl', cookie='', mock_connection=True):
        mysock = mock.Mock()
        mysock.gettimeout.return_value = 4242

        myresponse = mock.Mock()
        myresponse.read.return_value = 'body'
        myresponse.getheaders.return_value = dict([('head', '1')])
        myresponse.status = 301

        myconn = mock.Mock()
        myconn.request.return_value = None
        myconn.sock = mysock
        myconn.getresponse.return_value = myresponse
        myconn.__str__ = mock.Mock()
        myconn.__str__.return_value = 'myconn string'

        req = self.req
        req._redirect_params = mock.Mock()
        req._redirect_params.return_value = (myconn, url)
        req._request_str = mock.Mock()
        req._request_str.return_value = 'http://cool/cool'

        client = self.client
        client.need_login = False
        client._auto_login = False
        client._auth_cookie = False
        if mock_connection:
            client.acquire_connection = mock.Mock()
            client.acquire_connection.return_value = myconn
        client.release_connection = mock.Mock()
        client.set_auth_cookie = mock.Mock()
        client._wait_for_login = mock.Mock()
        client.auth_cookie = mock.Mock()
        client.auth_cookie.return_value = cookie
        return (mysock, myresponse, myconn)

    def test_issue_request_trigger_exception(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        self.client.acquire_connection.return_value = None

        self.req._issue_request()
        self.assertIsInstance(self.req._request_error, Exception)
        self.assertTrue(self.client.acquire_connection.called)

    def test_issue_request_handle_none_sock(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        myconn.sock = None
        self.req.start()
        self.assertIsNone(self.req.join())
        self.assertTrue(self.client.acquire_connection.called)

    def test_issue_request_exceed_maximum_retries(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        myresponse.status = 401
        self.req.start()
        self.assertIsNone(self.req.join())
        self.assertTrue(self.client.acquire_connection.called)

    def test_issue_request_relogin(self):
        (mysock, myresponse, myconn) = self.prep_issue_request(
            url='login?redir=%2fapi%2fv2')
        self.req.start()
        self.assertIsNone(self.req.join())
        self.assertTrue(self.client._wait_for_login.called)
        for args, kwargs in self.client.set_auth_cookie.call_args_list:
            if not args[1]:
                continue
            self.assertIsNotNone(args[1])

    def test_issue_request_with_cookie(self):
        (mysock, myresponse, myconn) = self.prep_issue_request(
            cookie=dict([('Cookie', 'ck'), ('X-CSRFTOKEN', 'tk')]))
        self.req.start()
        self.req.join()
        self.assertTrue(myconn.request.called)
        for args, kwargs in myconn.request.call_args_list:
            if "Cookie" not in args[3]:
                continue
            self.assertEqual(args[3]["Cookie"], 'ck')

    #TODO(jerryz): something didn't run
    def noop_test_issue_request_ok(self):
        (mysock, myresponse, myconn) = self.prep_issue_request(
            cookie=dict([('Cookie', 'ck'), ('X-CSRFTOKEN', 'tk')]))
        myresponse.status = 200
        self.req.start()
        self.assertIsNotNone(self.req.join())

    def test_acquire_connection_ok(self):
        (mysock, myresponse, myconn) = self.prep_issue_request(
            mock_connection=False)
        conn = self.client.acquire_connection(
            self, headers=None, rid=-1)
        self.assertIsInstance(
            conn, httplib.HTTPConnection)

    def test_issue_request_with_no_cookie(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        myresponse.status = 401
        self.req.start()
        self.assertIsNone(self.req.join())
        self.client.set_auth_cookie.assert_called_with(myconn, None)

    def test_issue_request_exceed_maximum_redirects(self):
        logger = logging.getLogger(
            'networking_fortinet.api_client.request')
        with mock.patch.object(logger, 'info') as mock_log:
            (mysock, myresponse, myconn) = self.prep_issue_request()
            self.req.start()
            self.assertIsNone(self.req.join())
            self.assertTrue(self.client.acquire_connection.called)
            for args, kwargs in mock_log.call_args_list:
                if "Maximum redirects exceeded" not in args:
                    continue
                self.assertTrue("Maximum redirects exceeded" in args)
            self.assertTrue(mock_log.called)

    def test_issue_request_refresh_cookie(self):
        (mysock, myresponse, myconn) = self.prep_issue_request(
            cookie=dict([('Cookie', 'old'), ('X-CSRFTOKEN', 'tk')]))
        myresponse.status = 401
        self.req.start()
        self.assertIsNone(self.req.join())
        self.client.set_auth_cookie.assert_called_with(myconn, None)

    def test_issue_request_unavailable(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        myresponse.status = 503
        self.req.start()
        self.assertIsNone(self.req.join())
        self.client.release_connection.called_with(
            myconn, True, True, mock.ANY)

#    def test_issue_request_internal_server_error(self):
#        (mysock, myresponse, myconn) = self.prep_issue_request()
#        myresponse.status = httplib2.INTERNAL_SERVER_ERROR
#        with self.assertRaises(Exception, 'Server error return: 500'):
#        self.req.start()
#        self.assertIsNone(self.req.join())

    def test_issue_request_trigger_non_redirect(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        myresponse.status = 200
        self.req.start()
        self.assertIsNone(self.req.join())
        self.assertTrue(self.client.acquire_connection.called)

    def test_issue_request_trigger_internal_server_error(self):
        (mysock, myresponse, myconn) = self.prep_issue_request()
        self.req._redirect_params.return_value = (myconn, None)
        self.req.start()
        self.assertIsNone(self.req.join())
        self.assertTrue(self.client.acquire_connection.called)

    def test_redirect_params_break_on_location(self):
        myconn = mock.Mock()
        (conn, retval) = self.req._redirect_params(
            myconn, [('location', None)])
        self.assertIsNone(retval)

    def test_redirect_params_parse_a_url(self):
        myconn = mock.Mock()
        (conn, retval) = self.req._redirect_params(
            myconn, [('location', '/path/a/b/c')])
        self.assertIsNotNone(retval)

    def test_redirect_params_invalid_redirect_location(self):
        myconn = mock.Mock()
        (conn, retval) = self.req._redirect_params(
            myconn, [('location', '+path/a/b/c')])
        self.assertIsNone(retval)

    def test_redirect_params_invalid_scheme(self):
        myconn = mock.Mock()
        (conn, retval) = self.req._redirect_params(
            myconn, [('location', 'invalidscheme://hostname:1/path')])
        self.assertIsNone(retval)

    def test_redirect_params_setup_https_with_cookie(self):
            self.req._api_client = self.fortiosclient
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(
                myconn, [('location', 'https://host:1/path')])

            self.assertIsNotNone(retval)
            self.assertIsInstance(conn, httplib.HTTPSConnection)

    def test_redirect_params_setup_http_with_cookie(self):
            self.req._api_client = self.fortiosclient
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(
                myconn, [('location', 'http://host:1/path')])

            self.assertIsNotNone(retval)
            self.assertIsInstance(conn, httplib.HTTPConnection)

    def test_redirect_params_setup_http_existing_pool(self):
            self.req._api_client = self.fortiosclient
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(
                myconn, [('location', 'http://127.0.0.1:80/abc')])

            self.assertIsNotNone(retval)
            self.assertIsInstance(conn, httplib.HTTPConnection)

    def test_redirect_params_setup_https_and_query(self):
        with mock.patch(EVT_CLIENT_PATH) as mock_client:
            api_client = mock_client.return_value
            self.req._api_client = api_client
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(myconn, [
                ('location', 'https://host:1/path?q=1')])

            self.assertIsNotNone(retval)
            self.assertTrue(api_client.acquire_redirect_connection.called)

    def test_redirect_params_setup_https_connection_no_cookie(self):
        with mock.patch(EVT_CLIENT_PATH) as mock_client:
            api_client = mock_client.return_value
            self.req._api_client = api_client
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(myconn, [
                ('location', 'https://host:1/path')])

            self.assertIsNotNone(retval)
            self.assertTrue(api_client.acquire_redirect_connection.called)

    def test_redirect_params_setup_https_and_query_no_cookie(self):
        with mock.patch(EVT_CLIENT_PATH) as mock_client:
            api_client = mock_client.return_value
            self.req._api_client = api_client
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(
                myconn, [('location', 'https://host:1/path?q=1')])
            self.assertIsNotNone(retval)
            self.assertTrue(api_client.acquire_redirect_connection.called)

    def test_redirect_params_path_only_with_query(self):
        with mock.patch(EVT_CLIENT_PATH) as mock_client:
            api_client = mock_client.return_value
            api_client.wait_for_login.return_value = None
            api_client.auth_cookie = None
            api_client.acquire_connection.return_value = True
            myconn = mock.Mock()
            (conn, retval) = self.req._redirect_params(myconn, [
                ('location', '/path?q=1')])
            self.assertIsNotNone(retval)

    def test_handle_request_auto_login(self):
        self.req._auto_login = True
        self.req._api_client = mock.Mock()
        self.req._api_client.need_login = True
        self.req._request_str = mock.Mock()
        self.req._request_str.return_value = 'http://cool/cool'
        self.req.spawn = mock.Mock()
        self.req._handle_request()

    def test_handle_request_auto_login_unauth(self):
        self.req._auto_login = True
        self.req._api_client = mock.Mock()
        self.req._api_client.need_login = True
        self.req._request_str = mock.Mock()
        self.req._request_str.return_value = 'http://cool/cool'

        import socket
        resp = httplib.HTTPResponse(socket.socket())
        resp.status = 401
        mywaiter = mock.Mock()
        mywaiter.wait = mock.Mock()
        mywaiter.wait.return_value = resp
        self.req.spawn = mock.Mock(return_value=mywaiter)
        self.req._handle_request()

    def test_construct_eventlet_get_api_providers_request(self):
        r = request.GetApiProvidersRequestEventlet(self.client)
        self.assertIsNotNone(r)

    def test_api_providers_none_api_providers(self):
        r = request.GetApiProvidersRequestEventlet(self.client)
        r.successful = mock.Mock(return_value=False)
        self.assertIsNone(r.api_providers())

    def test_api_providers_non_none_api_providers(self):
        r = request.GetApiProvidersRequestEventlet(self.client)
        r.value = mock.Mock()
        r.value.body = """{
          "results": [
            { "roles": [
              { "role": "api_provider",
                "listen_addr": "pssl:1.1.1.1:1" }]}]}"""
        r.successful = mock.Mock(return_value=True)
        self.assertIsNotNone(r.api_providers())

    def test_construct_eventlet_login_request(self):
        r = request.LoginRequestEventlet(self.fortiosclient, 'user',
            'password')
        self.assertIsNotNone(r)

    def test_session_cookie_session_cookie_retrieval(self):
        r = request.LoginRequestEventlet(self.fortiosclient, 'user',
            'password')
        r.successful = mock.Mock()
        r.successful.return_value = True
        r.value = mock.Mock()
        r.value.get_header = mock.Mock()
        r.value.get_header.return_value = 'cool'
        self.assertIsNotNone(r.session_cookie())

    def test_session_cookie_not_retrieved(self):
        r = request.LoginRequestEventlet(self.fortiosclient, 'user',
            'password')
        r.successful = mock.Mock()
        r.successful.return_value = False
        r.value = mock.Mock()
        r.value.get_header = mock.Mock()
        r.value.get_header.return_value = 'cool'
        self.assertIsNone(r.session_cookie())



# Copyright 2015 Fortinet Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
This module provides a decorator for adding the sigleton feature to classes.
Example usage:

   >>> @singleton
   ... class SpamaTon(object):
   ...     def __init__(self,  adjunct):
   ...         self.adjunct = adjunct
   ...
   ...     def spam_adjunct(self):
   ...         return self.adjunct
   ...
   >>> EggJunc = SpamaTon("Eggs")
   >>> HamJunc = SpamaTon("Ham")
   >>> print('Spam with ' + EggJunc.spam_adjunct())
   Spam with Ham

   >>> @singleton
   ... class SingleInitSpamaTon(object):
   ...     @ignore_subsequent
   ...     def __init__(self,  adjunct):
   ...         self.adjunct = adjunct
   ...
   ...     def spam_adjunct(self):
   ...         return self.adjunct
   ...
   >>> EggJunc = SingleInitSpamaTon("Eggs")
   >>> HamJunc = SingleInitSpamaTon("Ham")
   >>> print('Spam with ' + HamJunc.spam_adjunct())
   Spam with Eggs

For inheritance from a singleton (discouraged) __new__ needs to be overwritten
or the subclass will share its instance.
Example:
   >>> @singleton
   ... class SpamaTon(object):
   ...     def __init__(self,  adjunct):
   ...         self.adjunct = adjunct
   ...
   ...     def spam_adjunct(self):
   ...         return self.adjunct
   ...
   >>> @singleton
   ... class SubSpamaTon(SpamaTon):
   ...     def __new__(cls,  *args,  **kwargs):
   ...         return super(SpamaTon.__class__,  cls).__new__(cls)
   ...
   ...     def __init__(self,  adjunct,  second_adjunct):
   ...         super(SubSpamaTon,  self).__init__(adjunct)
   ...         self.second_adjunct = second_adjunct
   ...
   ...     def spam_adjunct(self):
   ...         return ' and '.join([self.adjunct,  self.second_adjunct])
   ...
   >>> MultiJunc = SubSpamaTon("Ham",  "Eggs")
   >>> print('Spam with ' + MultiJunc.spam_adjunct())
   Spam with Ham and Eggs
"""
import functools

try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading


class SingletonFactory(object):
    """Takes a class and produces a singleton."""

    def __init__(self, cls, lock):
        """Create a new SingletonFactory.
       Keyword arguments:
       cls -- class to produce singleton from
       lock -- Lock object shared between threads to synchronize instance
       access
       """
        self._lock = lock
        self._old_cls = cls
        self._instance = None

    def get_new(self):
        """Returns a replacement for the __new__ functions of the class to
       turn into a singleton. The replacement first will lock access for other
       threads. Then a new instance of the singleton is created. Further
       constructor calls will result in a reference to the single instance.
       The replacement function is wrapped to be transparent to any changes to
       __new__ in the original class.
       """

        @functools.wraps(self._old_cls.__new__)
        def new_instantiated(cls, *args, **kwargs):
            """Called to get an existing instance of a singleton.
           Keyword arguments:
           cls, args, kwargs -- ignored, exist for calling conventions only
           """
            return self._instance

        @functools.wraps(self._old_cls.__new__)
        def new_uninstantiated(cls, *args, **kwargs):
            """Called when a new singleton instance is created.
           This first locks access for other threads. Then a new instance of
           cls is created and further calls to __new__ are replaced by
           new_instantiated
           Keyword arguments:
           cls -- instantiated class
           args, kwargs -- optional arguments to pass to __init__ of the
           instatiated class
           """

            with self._lock:
                if not self._instance:
                    self._instance = self._old_cls.__new__(cls)
                    cls.__new__ = classmethod(new_instantiated)

            return self._instance

        return classmethod(new_uninstantiated)

    def __call__(self):
        """Produces a singleton.
       Returns a copy of the original class which has its __new__ method
       replaced by new_uninstantiated
       """
        new_cls = type(self._old_cls.__name__, self._old_cls.__bases__,
                       dict(self._old_cls.__dict__))
        new_cls.__new__ = self.get_new()
        return new_cls


def singleton(cls):
    """Decorates a class making it a singleton.
   Keyword arguments:
   cls -- class to decorate.
   """
    factory = SingletonFactory(cls, _threading.RLock())
    return factory()


def ignore_subsequent(instance_method):
    """Decorates an instance method to be ignored if called subsequently.
   Keyword arguments:
   instance_method -- instance method to be decorated
   """

    @functools.wraps(instance_method)
    def ignore_method(self, *args, **kwargs):
        pass

    @functools.wraps(instance_method)
    def first_method(self, *args, **kwargs):
        instance_method(self, *args, **kwargs)
        setattr(self.__class__, instance_method.__name__, ignore_method)

    return first_method


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# Copyright 2015 Fortinet, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import uuid

from eventlet import event
from eventlet import greenthread
from oslo_log import log as logging
from oslo_service import loopingcall
import six

from neutron.common import exceptions

from networking_fortinet.common import resources
from networking_fortinet.common import singleton
from networking_fortinet.tasks import constants


DEFAULT_INTERVAL = 1000

LOG = logging.getLogger(__name__)


def nop(task):
    return constants.TaskStatus.COMPLETED


class TaskException(exceptions.NeutronException):

    def __init__(self, message=None, **kwargs):
        if message is not None:
            self.message = message
        super(TaskException, self).__init__(**kwargs)


class InvalidState(TaskException):
    message = _("Invalid state %(state)d")


class TaskStateSkipped(TaskException):
    message = _("State %(state)d skipped. Current state %(current)d")


class Tasks(object):
    def __init__(self, id):
        #self.name = name if name else uuid.uuid1()
        #self.client = client
        # self._tasks is a stack to store the rollback tasks of tasks executed.
        # self._tasks example
        # [
        #   {
        #       'data': {'vdom': 'osvdm1'},
        #       'func': <function wrapper at 0x3ee2140>
        #   },
        #   {
        #       'data': {'vdom': 'osvdm1'},
        #       'func': <function wrapper at 0x3ee21b8>
        #   }
        # ]
        self._tasks = collections.deque()
        # task id should be unified, here we can use context.request_id
        self.id = id if id else str(uuid.uuid1())
        self.state = event.Event()
        self.status = constants.TaskStatus.NONE
        self.status_handlers = {
            constants.TaskStatus.ROLLBACK: self.execute,
            constants.TaskStatus.COMPLETED: self._finished,
            constants.TaskStatus.ERROR: self._finished,
            constants.TaskStatus.ABORT: self._finished
        }

    def register(self, **subtask):
        if subtask in self._tasks:
            return
        self._tasks.append(subtask)
        self._update_status(constants.TaskStatus.PENDING)
        return self

    def _reset_tasks(self, status=constants.TaskStatus.NONE):
        while len(self._tasks):
            self._tasks.pop()
        self.status = status

    def execute(self):
        if constants.TaskStatus.ROLLBACK != self.status:
            return
        while len(self._tasks):
            try:
                subtask = self._tasks.pop()
                subtask['func'](*subtask['params'])
            except Exception:
                msg = (_("Task %(task)s encountered exception in %(func)s "),
                       {'task': str(self), 'func': str(subtask['func'])})
                self.status = constants.TaskStatus.ERROR
                self._tasks = {}
                LOG.exception(msg)
        return self._update_status(constants.TaskStatus.COMPLETED)

    def _update_status(self, status):
        if status != self.status:
            self.status = status
            if status in self.status_handlers.keys():
                self.status_handlers[status]()
        return self.status

    def _finished(self):
        self._reset_tasks(constants.TaskStatus.COMPLETED)

    def wait(self):
        if constants.TaskStatus.NONE == self.status:
            return
        status = self.state.wait()
        self.status_handlers[status]()

    def __repr__(self):
        return "Task-%s" % (self.id)


@singleton.singleton
class TaskManager(object):

    _instance = None
    _default_interval = DEFAULT_INTERVAL

    def __init__(self, interval=None):
        self._interval = interval or TaskManager._default_interval

        # A queue to pass tasks from other threads
        self._tasks_queue = collections.deque()

        # A dict to task id
        self._tasks = {}

        # Current task being executed in main thread
        self._main_thread_exec_task = None

        # New request event
        self._req = event.Event()

        # TaskHandler stopped event
        self._stopped = False

        # Periodic function trigger
        self._monitor = None
        self._monitor_busy = False

        # Thread handling the task request
        self._thread = None

    def _execute(self, task):
        """Execute task."""
        msg = _("@@@ Start task %s") % str(task)
        LOG.debug(msg)
        try:
            task.wait()
        except Exception:
            msg = (_("Task %(task)s encountered exception"),
                {'task': str(task)})
            LOG.exception(msg)
            #status = constants.TaskStatus.ERROR
        LOG.debug("Task %(task)s return", {'task': str(task)})

    def _result(self, task):
        """Notify task execution result."""
        try:
            return
        except Exception:
            msg = _("Task %(task)s encountered exception in %(cb)s") % {
                'task': str(task),
                'cb': str(task._result_callback)}
            LOG.exception(msg)

        LOG.debug("Task %(task)s return %(status)s",
                  {'task': str(task), 'status': task.status})

        task._finished()

    def _check_pending_tasks(self):
        """Check all pending tasks status."""
        for id in self._tasks.keys():
            if self._stopped:
                # Task manager is stopped, return now
                return
            task = self._tasks[id]
            # only the first task is executed and pending
            if constants.TaskStatus.PENDING != task.status:
                self._dequeue(task)

    def _enqueue(self, id):
        if id not in self._tasks:
            self._tasks[id] = Tasks(id)
            self._tasks_queue.append(self._tasks[id])

    def _dequeue(self, task):
        if task in self._tasks_queue:
            self._tasks_queue.remove(task)
            del self._tasks[task.id]
            return

    def update_status(self, id, status):
        if id in self._tasks:
            self._tasks[id]._update_status(status)

    def _abort(self):
        """Abort all tasks."""
        # put all tasks haven't been received by main thread to queue
        # so the following abort handling can cover them
        for t in self._tasks_queue:
            self._enqueue(t)
        self._tasks_queue.clear()

        for id in self._tasks.keys():
            tasks = list(self._tasks[id])
            for task in tasks:
                task._update_status(constants.TaskStatus.ABORT)
                self._dequeue(task)

    def _get_task(self):
        """Get task request."""
        while True:
            for t in self._tasks_queue:
                if t.status in [constants.TaskStatus.ROLLBACK,
                                constants.TaskStatus.COMPLETED]:
                    return t
            self._req.wait()
            self._req.reset()

    def run(self):
        while True:
            try:
                if self._stopped:
                    # Gracefully terminate this thread if the _stopped
                    # attribute was set to true
                    LOG.info(_("Stopping TaskManager"))
                    break

                # get a task from queue, or timeout for periodic status check
                task = self._get_task()

                try:
                    #if constants.TaskStatus.ROLLBACK == task.status:
                    self._main_thread_exec_task = task
                    self._execute(task)
                finally:
                    self._main_thread_exec_task = None
                    if task.status in [constants.TaskStatus.NONE,
                                       constants.TaskStatus.ERROR,
                                       constants.TaskStatus.COMPLETED]:
                        # The thread is killed during _execute(). To guarantee
                        # the task been aborted correctly, put it to the queue.
                        #self._enqueue(task)
                        self._dequeue(task)
                    else:
                        self._enqueue(task)
            except Exception:
                LOG.exception(_("TaskManager terminating because "
                                "of an exception"))
                break

    def add(self, id, **subtask):
        if id is None:
            id = str(uuid.uuid1())
        if subtask:
            self._enqueue(id)
            self._tasks[id].register(**subtask)

    def stop(self):
        if self._thread is None:
            return
        self._stopped = True
        self._thread.kill()
        self._thread = None
        # Stop looping call and abort running tasks
        self._monitor.stop()
        if self._monitor_busy:
            self._monitor.wait()
        self._abort()
        LOG.info(_("TaskManager terminated"))

    def has_pending_task(self):
        if self._tasks_queue or self._tasks or self._main_thread_exec_task:
            return True
        else:
            return False

    def show_pending_tasks(self):
        for task in self._tasks_queue:
            LOG.info(str(task))
        for resource, tasks in six.iteritems(self._tasks):
            for task in tasks:
                LOG.info(str(task))
        if self._main_thread_exec_task:
            LOG.info(str(self._main_thread_exec_task))

    def count(self):
        count = 0
        for id, tasks in six.iteritems(self._tasks):
            count += len(tasks)
        return count

    def start(self, interval=None):
        def _inner():
            self.run()

        def _loopingcall_callback():
            self._monitor_busy = True
            try:
                self._check_pending_tasks()
            except Exception as e:
                resources.Exinfo(e)
                LOG.exception(_("Exception in _check_pending_tasks"))
            self._monitor_busy = False

        if self._thread is not None:
            return self

        if interval is None or interval == 0:
            interval = self._interval

        self._stopped = False
        self._thread = greenthread.spawn(_inner)
        self._monitor = loopingcall.FixedIntervalLoopingCall(
            _loopingcall_callback)
        self._monitor.start(interval / 1000.0,
                            interval / 1000.0)
        # To allow the created thread start running
        greenthread.sleep(0)
        return self

    @classmethod
    def set_default_interval(cls, interval):
        cls._default_interval = interval



# Copyright (c) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import exceptions as lib_exc

from tempest.lib.services.network import base


class FirewallsClient(base.BaseNetworkClient):

    def create_firewall(self, **kwargs):
        uri = '/fw/firewalls'
        post_data = {'firewall': kwargs}
        return self.create_resource(uri, post_data)

    def update_firewall(self, firewall_id, **kwargs):
        uri = '/fw/firewalls/%s' % firewall_id
        post_data = {'firewall': kwargs}
        return self.update_resource(uri, post_data)

    def show_firewall(self, firewall_id, **fields):
        uri = '/fw/firewalls/%s' % firewall_id
        return self.show_resource(uri, **fields)

    def delete_firewall(self, firewall_id):
        uri = '/fw/firewalls/%s' % firewall_id
        return self.delete_resource(uri)

    def list_firewalls(self, **filters):
        uri = '/fw/firewalls'
        return self.list_resources(uri, **filters)

    def is_resource_deleted(self, id):
        try:
            self.show_firewall(id)
        except lib_exc.NotFound:
            return True
        return False

    @property
    def resource_type(self):
        """Returns the primary type of resource this client works with."""
        return 'firewall'


class FirewallRulesClient(base.BaseNetworkClient):

    def create_firewall_rule(self, **kwargs):
        uri = '/fw/firewall_rules'
        post_data = {'firewall_rule': kwargs}
        return self.create_resource(uri, post_data)

    def update_firewall_rule(self, firewall_rule_id, **kwargs):
        uri = '/fw/firewall_rules/%s' % firewall_rule_id
        post_data = {'firewall_rule': kwargs}
        return self.update_resource(uri, post_data)

    def show_firewall_rule(self, firewall_rule_id, **fields):
        uri = '/fw/firewall_rules/%s' % firewall_rule_id
        return self.show_resource(uri, **fields)

    def delete_firewall_rule(self, firewall_rule_id):
        uri = '/fw/firewall_rules/%s' % firewall_rule_id
        return self.delete_resource(uri)

    def list_firewall_rules(self, **filters):
        uri = '/fw/firewall_rules'
        return self.list_resources(uri, **filters)


class FirewallPoliciesClient(base.BaseNetworkClient):

    def create_firewall_policy(self, **kwargs):
        uri = '/fw/firewall_policies'
        post_data = {'firewall_policy': kwargs}
        return self.create_resource(uri, post_data)

    def update_firewall_policy(self, firewall_policy_id, **kwargs):
        uri = '/fw/firewall_policies/%s' % firewall_policy_id
        post_data = {'firewall_policy': kwargs}
        return self.update_resource(uri, post_data)

    def show_firewall_policy(self, firewall_policy_id, **fields):
        uri = '/fw/firewall_policies/%s' % firewall_policy_id
        return self.show_resource(uri, **fields)

    def delete_firewall_policy(self, firewall_policy_id):
        uri = '/fw/firewall_policies/%s' % firewall_policy_id
        return self.delete_resource(uri)

    def list_firewall_policies(self, **filters):
        uri = '/fw/firewall_policies'
        return self.list_resources(uri, **filters)

    def insert_firewall_rule_in_policy(self, firewall_policy_id,
                                       firewall_rule_id, insert_after='',
                                       insert_before=''):
        uri = '/fw/firewall_policies/%s/insert_rule' % firewall_policy_id
        data = {
            'firewall_rule_id': firewall_rule_id,
            'insert_after': insert_after,
            'insert_before': insert_before,
        }
        return self.update_resource(uri, data)

    def remove_firewall_rule_from_policy(self, firewall_policy_id,
                                         firewall_rule_id):
        uri = '/fw/firewall_policies/%s/remove_rule' % firewall_policy_id
        data = {
            'firewall_rule_id': firewall_rule_id,
        }
        return self.update_resource(uri, data)



# Copyright (c) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from tempest import exceptions
from tempest.lib.common import ssh
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc
from tempest.scenario import manager

from networking_fortinet.tests.tempest_plugin.tests import fwaas_client

CONF = cfg.CONF


class FWaaSScenarioTest(fwaas_client.FWaaSClientMixin,
                        manager.NetworkScenarioTest):

    def check_connectivity(self, ip_address, username=None, private_key=None,
                           should_connect=True,
                           check_icmp=True, check_ssh=True,
                           check_reverse_icmp_ip=None,
                           should_reverse_connect=True,
                           check_reverse_curl=False):
        # if default allow is enabled as default by fgt fwaas, reverse
        # connection should always be available.
        if self._default_allow():
            should_reverse_connect = True
        if should_connect:
            msg = "Timed out waiting for %s to become reachable" % ip_address
        else:
            msg = "ip address %s is reachable" % ip_address
        if check_icmp:
            ok = self.ping_ip_address(ip_address,
                                      should_succeed=should_connect)
            self.assertTrue(ok, msg=msg)
        if check_ssh:
            connect_timeout = CONF.validation.connect_timeout
            kwargs = {}
            if not should_connect:
                # Use a shorter timeout for negative case
                kwargs['timeout'] = 1
            try:
                client = ssh.Client(ip_address, username, pkey=private_key,
                                    channel_timeout=connect_timeout,
                                    **kwargs)
                client.test_connection_auth()
                self.assertTrue(should_connect, "Unexpectedly reachable")
                if check_reverse_icmp_ip:
                    cmd = 'ping -c1 -w2 %s' % check_reverse_icmp_ip
                    try:
                        client.exec_command(cmd)
                        self.assertTrue(should_reverse_connect,
                                        "Unexpectedly reachable (reverse)")
                    except lib_exc.SSHExecCommandFailed:
                        if should_reverse_connect:
                            raise
                    if check_reverse_curl:
                        cmd1 = 'curl http://httpstat.us/200 |grep "200 OK"'
                        cmd2 = 'curl http://www.eicar.org/download/eicar.com|\
                                grep EICAR-STANDARD-ANTIVIRUS-TEST-FILE'
                        try:
                            client.exec_command(cmd1)
                            self.assertTrue(should_reverse_connect,
                                            "Unexpectedly reachable (reverse)")
                        except lib_exc.SSHExecCommandFailed:
                            if should_reverse_connect:
                                raise
                        # test virus file download should be blocked by default
                        # security profile enabled.
                        try:
                            client.exec_command(cmd2)
                            self.assertFalse(should_reverse_connect,
                                            "Unexpectedly reachable (reverse)")
                            raise
                        except lib_exc.SSHExecCommandFailed:
                            if should_reverse_connect:
                                pass
            except lib_exc.SSHTimeout:
                if should_connect:
                    raise

    def create_networks(self, networks_client=None,
                        routers_client=None, subnets_client=None,
                        tenant_id=None, dns_nameservers=None):
        """Create a network with a subnet connected to a router.
        The baremetal driver is a special case since all nodes are
        on the same shared network.
        :param client: network client to create resources with.
        :param tenant_id: id of tenant to create resources in.
        :param dns_nameservers: list of dns servers to send to subnet.
        :returns: network, subnet, router
        """
        if CONF.baremetal.driver_enabled:
            # NOTE(Shrews): This exception is for environments where tenant
            # credential isolation is available, but network separation is
            # not (the current baremetal case). Likely can be removed when
            # test account mgmt is reworked:
            # https://blueprints.launchpad.net/tempest/+spec/test-accounts
            if not CONF.compute.fixed_network_name:
                m = 'fixed_network_name must be specified in config'
                raise exceptions.InvalidConfiguration(m)
            network = self._get_network_by_name(
                CONF.compute.fixed_network_name)
            router = None
            subnet = None
        else:
            network = self._create_network(
                networks_client=networks_client,
                tenant_id=tenant_id)
            router = self._get_router(client=routers_client,
                                      tenant_id=tenant_id)

            subnet_kwargs = dict(network=network,
                                 subnets_client=subnets_client,
                                 routers_client=routers_client)
            # use explicit check because empty list is a valid option
            if dns_nameservers is not None:
                subnet_kwargs['dns_nameservers'] = dns_nameservers
            subnet = self._create_subnet(**subnet_kwargs)
            if not routers_client:
                routers_client = self.routers_client
            routers_client.add_router_interface(router['id'],
                                                subnet_id=subnet['id'])
            # save a cleanup job to remove this association between
            # router and subnet
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            routers_client.remove_router_interface,
                            router['id'], subnet_id=subnet['id'])
        return network, subnet, router

    def _get_router(self, client=None, tenant_id=None):
        """Retrieve a router for the given tenant id.
        If a public router has been configured, it will be returned.
        If a public router has not been configured, but a public
        network has, a tenant router will be created and returned that
        routes traffic to the public network.
        """
        if not client:
            client = self.routers_client
        if not tenant_id:
            tenant_id = client.tenant_id
        router_id = CONF.network.public_router_id
        network_id = CONF.network.public_network_id
        if router_id:
            body = client.show_router(router_id)
            return body['router']
        elif network_id:
            # fortigate plugin only allow one router per tenant, so if
            # a router already exists, use it.
            routers_list = client.list_routers(tenant_id=tenant_id)
            if len(routers_list['routers']) == 1:
                router = routers_list['routers'][0]
            else:
                router = self._create_router(client, tenant_id)
            kwargs = {'external_gateway_info': dict(network_id=network_id)}
            router = client.update_router(router['id'], **kwargs)['router']
            return router
        else:
            raise Exception("Neither of 'public_router_id' or "
                            "'public_network_id' has been defined.")

    def _default_allow(self):
        if CONF.fortigate.enable_default_fwrule:
            return False
        else:
            return True



# Copyright 2015 Fortinet Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import os
from oslo_log import log as logging
import re
import six
import sys
import types

from networking_fortinet.api_client import exception as api_ex
from networking_fortinet.common import constants as const


LOG = logging.getLogger(__name__)

OPS = ["ADD", "DELETE", "SET", "GET", "MOVE"]
RB_FUNC = {'add': 'delete'}


# For debug purpose
def funcinfo(cls=None, action=None, data=None):
    cur_func = inspect.stack()[1][3]
    caller = inspect.stack()[2][3]
    LOG.debug("## current function is %(cur_func)s,"
              "its caller is %(caller)s",
              {'cur_func': cur_func, 'caller': caller})
    if cls or action or data:
        LOG.debug("## cls: %(cls)s, action: %(action)s, data: %(data)s",
                  {'cls': cls.__name__, 'action': action, 'data': data})


def rollback(func):
    def wrapper(cls, *args):
        result = func(cls, *args)
        if not result:
            rollback = {}
        else:
            rollback = cls._prepare_rollback(cls.delete, *args, **result)
        return {'result': result, 'rollback': rollback}
    return wrapper


class Exinfo(object):
    def __init__(self, exception):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        LOG.error(_("An exception of type %(exception)s occured with "
                    "arguments %(args)s, line %(line)s, in %(file)s"),
                  {'exception': type(exception).__name__,
                   'args': exception.args,
                   'line': exc_tb.tb_lineno,
                   'file': fname})


class DefaultClassMethods(type):
    def __getattr__(cls, attr):
        if str(attr).upper() not in OPS:
            raise AttributeError(attr)
        if 'ADD' == str(attr).upper():
            @rollback
            def _defaultClassMethod(cls, client, data):
                return cls.element(client, attr, data)
        else:
            def _defaultClassMethod(cls, client, data):
                return cls.element(client, attr, data)
        return types.MethodType(_defaultClassMethod, cls)


@six.add_metaclass(DefaultClassMethods)
class Base(object):
    def __init__(self):
        self.exist = False
        self.rollback = None

    @staticmethod
    def params_decoded(*args):
        keys = ['client', 'data']
        return dict(zip(keys, args))

    @classmethod
    def _prepare_rollback(cls, func, *args, **result):
        if not func:
            return None
        params = cls.params_decoded(*args)
        data = cls._rollback_data(params, **result)
        rollback = {
            'func': func,
            'params': (params['client'], data)
        }
        return rollback

    @classmethod
    def _rollback_data(cls, params, **result):
        return {
            'vdom': params['data'].get('vdom', const.EXT_VDOM),
            'name': params['data']['name']
        }

    @classmethod
    def element(cls, client, action, data):
        funcinfo(cls=cls, action=action, data=data)
        if not data:
            data = getattr(cls, 'data', None)
        # op is the combination of action and resource class name,
        # all ops should be defined in the templates
        name = re.findall("[A-Z][^A-Z]*", cls.__name__)
        op = "%s_%s" % (str(action).upper(), "_".join(name).upper())
        try:
            return client.request(op, **data)
        except api_ex.ApiException as e:
            Exinfo(e)
            raise e


class Vdom(Base):
    def __init__(self):
        super(Vdom, self).__init__()

    @classmethod
    def _rollback_data(cls, params, **result):
        return {'name': params['data'].get('name')}


class VdomLink(Base):
    def __init__(self):
        super(VdomLink, self).__init__()


class VlanInterface(Base):
    def __init__(self):
        super(VlanInterface, self).__init__()

"""
    @classmethod
    def get_with_params_check(cls, driver, data):
        res = cls.get(driver, data)
        for key, value in data.iteritems():
            if value != res['results'][0].get(key):
                res = cls.set(driver, data)
                break
        return res
"""


class RouterStatic(Base):
    def __init__(self):
        super(RouterStatic, self).__init__()

    @classmethod
    def _rollback_data(cls, params, **result):
        return {
            'vdom': params['data']['vdom'],
            'id': result['results']['mkey']
        }


class FirewallIppool(Base):
    def __init__(self):
        super(FirewallIppool, self).__init__()


class FirewallPolicy(Base):
    def __init__(self):
        super(FirewallPolicy, self).__init__()

    @classmethod
    def _rollback_data(cls, params, **result):
        return {
            'vdom': params['data']['vdom'],
            'id': result['results']['mkey']
        }


class FirewallAddress(Base):
    def __init__(self):
        super(FirewallAddress, self).__init__()


class FirewallAddrgrp(Base):
    def __init__(self):
        super(FirewallAddrgrp, self).__init__()


class FirewallService(Base):
    def __init__(self):
        super(FirewallService, self).__init__()


class DhcpServer(Base):
    def __init__(self):
        super(DhcpServer, self).__init__()

    @classmethod
    def _rollback_data(cls, params, **result):
        return {
            'vdom': params['data']['vdom'],
            'id': result['results']['mkey']
        }


class DhcpServerRsvAddr(Base):
    def __init__(self):
        super(DhcpServerRsvAddr, self).__init__()


class FirewallVip(Base):
    def __init__(self):
        super(FirewallVip, self).__init__()
